import numpy as np
from pyspedas.tplot_tools import get_data, store_data, options


def mms_fpi_make_errorflagbars(tname, level='l2'):
    """
    This procedure creates FPI error flag bars for plotting

        For DES/DIS moments:
            tname+'_flagbars_full': Detailed flag bars (all bars)
            tname+'_flagbars_main': Standard flag bars (4 bars)
            tname+'_flagbars_mini': Smallest flag bar (1 bars)

        For DES/DIS distribution functions
            tname+'_flagbars_dist': Standard flag bars (2 bars)

    Input
    -----------
        tname: str
            tplot variable containing the DIS or DES errorflag data

    Parameters
    -----------
        level: str
            Level of the data to create the errorflags bar for;
            (default is l2; this only needs to be set for QL data)

    Examples
    -----------
       > mms_fpi_make_errorflagbars('mms1_des_errorflags_fast')
       > mms_fpi_make_errorflagbars('mms1_dis_errorflags_fast')
       > mms_fpi_make_errorflagbars('mms1_des_errorflags_brst')
       > mms_fpi_make_errorflagbars('mms1_dis_errorflags_brst')

       For DES/DIS distribution function (Brst and Fast):
         bit 0 = manually flagged interval --> contact the FPI team for direction when utilizing this data; further correction is required
         bit 1 = overcounting/saturation effects likely present in skymap

       For DES/DIS moments (Brst):
         bit 0 = manually flagged interval  --> contact the FPI team for direction when utilizing this data; further correction is required
         bit 1 = overcounting/saturation effects likely present in skymap
         bit 2 = reported spacecraft potential above 20V
         bit 3 = invalid/unavailable spacecraft potential
         bit 4 = significant (>10%) cold plasma (<10eV) component
         bit 5 = significant (>25%) hot plasma (>30keV) component
         bit 6 = high sonic Mach number (v/vth > 2.5)
         bit 7 = low calculated density (n_DES < 0.05 cm^-3)
         bit 8 = onboard magnetic field used instead of brst l2pre magnetic field
         bit 9 = srvy l2pre magnetic field used instead of brst l2pre magnetic field
         bit 10 = no internally generated photoelectron correction applied
         bit 11 = compression pipeline error
         Bit 12 = spintone calculation error (DBCS only)
         Bit 13 = significant (>=20%) penetrating radiation (DIS only)
         Bit 14 = high MMS3 spintone due to DIS008 anomaly (DIS only)

       For DES/DIS moments (Fast):
         bit 0 = manually flagged interval --> contact the FPI team for direction when utilizing this data; further correction is required
         bit 1 = overcounting/saturation effects likely present in skymap
         bit 2 = reported spacecraft potential above 20V
         bit 3 = invalid/unavailable spacecraft potential
         bit 4 = significant (>10%) cold plasma (<10eV) component
         bit 5 = significant (>25%) hot plasma (>30keV) component
         bit 6 = high sonic Mach number (v/vth > 2.5)
         bit 7 = low calculated density (n_DES < 0.05 cm^-3)
         bit 8 = onboard magnetic field used instead of srvy l2pre magnetic field
         bit 9 = not used
         bit 10 = no internally generated photoelectron correction applied
         bit 11 = compression pipeline error
         Bit 12 = spintone calculation error (DBCS only)
         Bit 13 = significant (>=20%) penetrating radiation (DIS only)
         Bit 14 = high MMS3 spintone due to DIS008 anomaly (DIS only)

    Notes
    -----------
        EXPERIMENTAL!
        Based on the IDL version by Naritoshi Kitamura

    Returns
    -----------
        List containing the names of the created tplot variables
    """
    instrument = tname.split('_')[1].upper()
    data_rate = tname.split('_')[3].capitalize()

    data = get_data(tname, dt=True)
    metadata = get_data(tname, metadata=True)

    if metadata is None:
        return


    # Workaround for cdflib globalattsget() bug
    md_dt = metadata['CDF']['GATT']['Data_type']
    if isinstance(md_dt, list):
        md_dt = md_dt[0]

    if md_dt[-4:] == 'moms' or level == 'ql':
        labels_full=['Contact FPI team','Saturation','SCpot>20V','no SCpot','>10% Cold','>25% Hot','High Mach#','Low Density','Onboard Mag','L2pre Mag','Photoelectrons','Compression', 'Spintones', 'Radiation', 'MMS3 DIS Spintone']

        flags = ['{0:015b}'.format(flag) for flag in data.y]
        flagline = np.zeros((len(data.times), 15))
        flagline_others = np.zeros(len(data.times))
        flagline_all = np.zeros(len(data.times))
        for j, flag in enumerate(flags):
            for i in range(15):
                if int(flag[14-i:14-i+1]) == 0:
                    flagline[j, i] = np.nan
                    if flagline_all[j] != 1:
                        flagline_all[j] = np.nan
                    if instrument == 'DES':
                        if i != 1 and i != 4 and i != 5:
                            if flagline_others[j] != 1:
                                flagline_others[j] = np.nan
                    else:
                        if i != 1 and i != 5 and i != 13:
                            if flagline_others[j] != 1:
                                flagline_others[j] = np.nan
                else:
                    if instrument == 'DES':
                        if i != 1 and i != 4 and i != 5:
                            flagline_others[j] = 1
                    else:
                        if i != 1 and i != 5 and i != 13:
                            flagline_others[j] = 1
                        if i != 1 and i != 4 and i != 5:
                            flagline_others[j] = 1
                    flagline_all[j] = 1.0
                    flagline[j, i] = 1.0

        labels_full=['Contact FPI team','Saturation','SCpot>20V','no SCpot','>10% Cold','>25% Hot','High Mach#','Low Density','Onboard Mag','L2pre Mag','Photoelectrons','Compression', 'Spintones', 'Radiation', 'MMS3 DIS Spintone']

        if instrument == 'DES':
            des_full = np.array([[flagline[:,0]],[flagline[:,1]-0.1],[flagline[:,2]-0.2],[flagline[:,3]-0.3],[flagline[:,4]-0.4],[flagline[:,5]-0.5],[flagline[:,6]-0.6],[flagline[:,7]-0.7],[flagline[:,8]-0.8],[flagline[:,9]-0.9],[flagline[:,10]-1.0],[flagline[:,11]-1.1],[flagline[:,12]-1.2]])
            des_full = des_full.squeeze().T
            store_data(tname + '_flagbars_full', data={'x': data.times, 'y': des_full})
            options(tname + '_flagbars_full', 'yrange', [-0.15, 1.25])
            options(tname + '_flagbars_full', 'legend_names', labels_full)
            options(tname + '_flagbars_full', 'symbols', True)
            options(tname + '_flagbars_full', 'markers', 's')
            options(tname + '_flagbars_full', 'thick', 3)
            options(tname + '_flagbars_full', 'panel_size', 0.8)
            options(tname + '_flagbars_full', 'border', False)
            options(tname + '_flagbars_full', 'ytitle', instrument)
            options(tname + '_flagbars_full', 'ysubtitle', data_rate)
            options(tname + '_flagbars_full', 'color', ['k', 'red', 'green', 'teal', 'blue', 'purple', 'teal', 'black', 'blue', 'green', 'red', 'k', 'blue'])

            des_main = np.array([[flagline[:,1]-0.2],[flagline[:,4]-0.4],[flagline[:,5]-0.6],[flagline_others-0.8]])
            des_main = des_main.squeeze().T
            store_data(tname + '_flagbars_main', data={'x': data.times, 'y': des_main})
            options(tname + '_flagbars_main', 'yrange', [0.1, 0.9])
            options(tname + '_flagbars_main', 'color', ['red', 'blue', 'purple', 'black'])
            options(tname + '_flagbars_main', 'legend_names', ['Saturation','Cold (>10%)','Hot (>25%)','Others'])
            options(tname + '_flagbars_main', 'ytitle', instrument)
            options(tname + '_flagbars_main', 'ysubtitle', data_rate)
            options(tname + '_flagbars_main', 'border', False)
            options(tname + '_flagbars_main', 'symbols', True)
            options(tname + '_flagbars_main', 'markers', 's')
            options(tname + '_flagbars_main', 'thick', 3)

            des_others = np.array([[flagline[:,11]-0.8],[flagline[:,12]-0.8],[flagline[:,7]-0.8],[flagline[:,6]-0.8],[flagline[:,8]-0.8],[flagline[:,9]-0.8],[flagline[:,2]-0.8],[flagline[:,3]-0.8],[flagline[:,10]-0.8],[flagline[:,0]-0.8]])
            des_others = des_others.squeeze().T
            store_data(tname + '_flagbars_others', data={'x': data.times, 'y': des_others})
            options(tname + '_flagbars_others', 'ytitle', instrument)
            options(tname + '_flagbars_others', 'ysubtitle', data_rate)
            options(tname + '_flagbars_others', 'border', False)
            options(tname + '_flagbars_others', 'thick', 4)

            store_data(tname + '_flagbars_mini', data={'x': data.times, 'y': flagline_all})
            options(tname + '_flagbars_mini', 'ytitle', instrument)
            options(tname + '_flagbars_mini', 'ysubtitle', data_rate)
            options(tname + '_flagbars_mini', 'yrange', [0.9, 1.1])
            options(tname + '_flagbars_mini', 'color', 'black')
            options(tname + '_flagbars_mini', 'panel_size', 0.1)
            options(tname + '_flagbars_mini', 'border', False)
            options(tname + '_flagbars_mini', 'thick', 4)
            options(tname + '_flagbars_mini', 'legend_names', 'Flagged')
            options(tname + '_flagbars_mini', 'symbols', True)
            options(tname + '_flagbars_mini', 'markers', 's')
        else:
            dis_full = np.array([[flagline[:,0]],[flagline[:,1]-0.1],[flagline[:,2]-0.2],[flagline[:,3]-0.3],[flagline[:,4]-0.4],[flagline[:,5]-0.5],[flagline[:,6]-0.6],[flagline[:,7]-0.7],[flagline[:,8]-0.8],[flagline[:,9]-0.9],[flagline[:,10]-1.0],[flagline[:,11]-1.1],[flagline[:,12]-1.2],[flagline[:,13]-1.3],[flagline[:,14]-1.4]])
            dis_full = dis_full.squeeze().T
            store_data(tname + '_flagbars_full', data={'x': data.times, 'y': dis_full})
            options(tname + '_flagbars_full', 'yrange', [-1.5, 1.35])
            options(tname + '_flagbars_full', 'legend_names', labels_full)
            options(tname + '_flagbars_full', 'symbols', True)
            options(tname + '_flagbars_full', 'markers', 's')
            options(tname + '_flagbars_full', 'thick', 3)
            options(tname + '_flagbars_full', 'panel_size', 0.8)
            options(tname + '_flagbars_full', 'border', False)
            options(tname + '_flagbars_full', 'ytitle', instrument)
            options(tname + '_flagbars_full', 'ysubtitle', data_rate)
            options(tname + '_flagbars_full', 'color', ['k', 'red', 'green', 'teal', 'blue', 'purple', 'teal', 'black', 'blue', 'green', 'red', 'k', 'blue', 'orange', 'red'])

            dis_main = np.array([[flagline[:,1]-0.2],[flagline[:,13]-0.4],[flagline[:,5]-0.6],[flagline_others-0.8]])
            dis_main = dis_main.squeeze().T
            store_data(tname + '_flagbars_main', data={'x': data.times, 'y': dis_main})
            options(tname + '_flagbars_main', 'yrange', [0.1, 0.9])
            options(tname + '_flagbars_main', 'color', ['red', 'orange', 'purple', 'black'])
            options(tname + '_flagbars_main', 'legend_names', ['Saturation','Radiation','Hot (>25%)','Others'])
            options(tname + '_flagbars_main', 'ytitle', instrument)
            options(tname + '_flagbars_main', 'ysubtitle', data_rate)
            options(tname + '_flagbars_main', 'border', False)
            options(tname + '_flagbars_main', 'symbols', True)
            options(tname + '_flagbars_main', 'markers', 's')
            options(tname + '_flagbars_main', 'thick', 3)

            dis_others = np.array([[flagline[:,11]-0.8],[flagline[:,12]-0.8],[flagline[:,7]-0.8],[flagline[:,6]-0.8],[flagline[:,8]-0.8],[flagline[:,9]-0.8],[flagline[:,2]-0.8],[flagline[:,3]-0.8],[flagline[:,10]-0.8],[flagline[:,0]-0.8]])
            dis_others = dis_others.squeeze().T
            store_data(tname + '_flagbars_others', data={'x': data.times, 'y': dis_others})
            options(tname + '_flagbars_others', 'ytitle', instrument)
            options(tname + '_flagbars_others', 'ysubtitle', data_rate)
            options(tname + '_flagbars_others', 'border', False)
            options(tname + '_flagbars_others', 'thick', 4)

            store_data(tname + '_flagbars_mini', data={'x': data.times, 'y': flagline_all})
            options(tname + '_flagbars_mini', 'ytitle', instrument)
            options(tname + '_flagbars_mini', 'ysubtitle', data_rate)
            options(tname + '_flagbars_mini', 'yrange', [0.9, 1.1])
            options(tname + '_flagbars_mini', 'color', 'black')
            options(tname + '_flagbars_mini', 'panel_size', 0.1)
            options(tname + '_flagbars_mini', 'border', False)
            options(tname + '_flagbars_mini', 'thick', 4)
            options(tname + '_flagbars_mini', 'legend_names', 'Flagged')
            options(tname + '_flagbars_mini', 'symbols', True)
            options(tname + '_flagbars_mini', 'markers', 's')

        out_vars = [tname + '_flagbars_full',
                    tname + '_flagbars_main',
                    tname + '_flagbars_others',
                    tname + '_flagbars_mini']
    elif md_dt[-4:] == 'dist':
        flags = ['{0:014b}'.format(flag) for flag in data.y]
        flagline = np.zeros((len(data.times), 2))
        for i in [0, 1]:
            for j in range(len(flags)):
                try:
                    flagset = int(flags[13-i:13-i+1][0])
                except IndexError:
                    continue
                if flagset == 0:
                    flagline[j, i] = np.nan
                else:
                    flagline[j, i] = 1

        dist_full = np.array([[flagline[:,0]-0.25],[flagline[:,1]-0.75]])
        dist_full = dist_full.squeeze().T
        store_data(tname + '_flagbars_dist', data={'x': data.times, 'y': dist_full})
        options(tname + '_flagbars_dist', 'yrange', [0, 1])
        options(tname + '_flagbars_dist', 'color', ['black', 'red'])
        options(tname + '_flagbars_dist', 'legend_names', ['Manually flagged', 'Saturation'])
        options(tname + '_flagbars_dist', 'panel_size', 0.25)
        options(tname + '_flagbars_dist', 'border', False)
        options(tname + '_flagbars_dist', 'thick', 4)
        options(tname + '_flagbars_dist', 'symbols', True)
        options(tname + '_flagbars_dist', 'markers', 's')
        out_vars = [tname + '_flagbars_dist']

    return out_vars



import logging
import pathlib
import importlib.resources
from typing import List, Dict

import numpy as np
from pyspedas.tplot_tools import time_double, time_string
from pyspedas.tplot_tools import get_timespan, store, get, store_data, get_data


def read_epde_calibration_data(path: pathlib.Path) -> List[Dict]:
    """
    Read ELFIN EPDE calibration data from file and return
    list of parsed calibration datasets.

    Parameters
    ----------
        path : pathlib.Path
            The path to the calibration data file.

    Returns
    ----------
        List of EPDE calibration data parsed from the file.

    """

    def parse_float_line(line: str):
        array_str = line.split(":")[1].strip()
        return [float(f.rstrip(".")) for f in array_str.split(",")]

    lines = []
    with open(path, "r") as f:
        lines = f.readlines()

    date_lines = (i for i, line in enumerate(lines) if line.startswith("Date"))
    calibrations = []
    for i in date_lines:

        date_str = lines[i].split(":")[1].strip()
        gf_str = lines[i+1].split(":")[1].strip()

        cal = {
            "date": time_double(date_str),
            "gf": float(gf_str),
            "overaccumulation_factors": parse_float_line(lines[i+2]),
            "thresh_factors": parse_float_line(lines[i+3]),
            "ch_efficiencies": parse_float_line(lines[i+4]),
            "ebins": parse_float_line(lines[i+5]),
        }

        calibrations.append(cal)

    return calibrations


def get_epde_calibration(calibration_data: Dict) -> Dict:
    """
    Performs basic transformations on calibration parameters loaded
    from a data file, e.g. calculate channel factors and create energy bin labels.

    Returned dictionary values are (mostly) numpy arrays for later usage.

    Parameters
    ----------
        calibration_data : dict
            A single set of calibration data read from a file. See `obj`:read_epde_calibration_data:

    Returns
    ----------
        Dictionary with EPD gain factor, overaccumulation factors, threshold factors,
        channel efficiencies, energy bins (and logmean), calibration channel factors,
        and the labels for the energy bins.

    """

    gf = calibration_data["gf"]

    overaccumulation_factors = np.array(calibration_data["overaccumulation_factors"])
    thresh_factors = np.array(calibration_data["thresh_factors"])
    ch_efficiencies = np.array(calibration_data["ch_efficiencies"])
    ebins = np.array(calibration_data["ebins"])

    cal_ch_factors = 1.0 / (gf * thresh_factors * ch_efficiencies)

    ebins_logmean = ebins.copy()
    for i in range(0, 15):
        i_mean_of_logs = (np.log10(ebins[i]) + np.log10(ebins[i+1])) / 2.0
        ebins_logmean[i] = np.power(10.0, i_mean_of_logs)
    ebins_logmean[15] = 6500.0

    ebin_labels = [
        "50-80", "80-120", "120-160", "160-210",
        "210-270", "270-345", "345-430", "430-630",
        "630-900", "900-1300", "1300-1800", "1800-2500",
        "2500-3350", "3350-4150", "4150-5800", "5800+",
    ]

    calibration = {
        "epd_gf": gf,
        "epd_overaccumulation_factors": overaccumulation_factors,
        "epd_thresh_factors": thresh_factors,
        "epd_ch_efficiencies": ch_efficiencies,
        "epd_ebins": ebins,
        "epd_cal_ch_factors": cal_ch_factors,
        "epd_ebins_logmean": ebins_logmean,
        "epd_ebin_labels": ebin_labels,
    }

    return calibration


def get_epdi_calibration() -> Dict:
    """
    Ported from IDL SPEDAS elf_get_epd_calibration procedure. There are
    no calibration data files for epdi. Current as of 2023-06-11.

    Vassilis Angelopoulos' most recent comments from IDL procedure reproduced.

    Returns
    ----------
        Dictionary with EPD gain factor, overaccumulation factors, threshold factors,
        channel efficiencies, energy bins (and logmean), calibration channel factors,
        and the labels for the energy bins.

    """

    # 9deg x 9deg (FWHM Solid Angle) by 1/4" dia. round aperture area
    gf = 0.031256

    overaccumulation_factors = np.ones(16)
    thresh_factors = np.ones(16)
    thresh_factors[0] = 0.2
    ch_efficiencies = np.ones(16)

    cal_ch_factors = 1.0 / (gf * thresh_factors * ch_efficiencies)

    # in keV based on Jiang Liu's Geant4 code 2019-3-5
    ebins = np.array([
        50.0, 80.0, 120.0, 160.0,
        210.0, 270.0, 345.0, 430.0,
        630.0, 900.0, 1300.0, 1800.0,
        2500.0, 3350.0, 4150.0, 5800.0,
    ])

    ebins_logmean = ebins.copy()
    for i in range(0, 15):
        i_mean_of_logs = (np.log10(ebins[i]) + np.log10(ebins[i+1])) / 2.0
        ebins_logmean[i] = np.power(10.0, i_mean_of_logs)
    ebins_logmean[15] = 6500.0

    # used same bins as electrons (dead layer well below 50keV)
    ebin_labels = [
        "50-80", "80-120", "120-160", "160-210",
        "210-270", "270-345", "345-430", "430-630",
        "630-900", "900-1300", "1300-1800", "1800-2500",
        "2500-3350", "3350-4150", "4150-5800", "5800+",
    ]

    calibration = {
        "epd_gf": gf,
        "epd_overaccumulation_factors": overaccumulation_factors,
        "epd_thresh_factors": thresh_factors,
        "epd_ch_efficiencies": ch_efficiencies,
        "epd_ebins": ebins,
        "epd_cal_ch_factors": cal_ch_factors,
        "epd_ebins_logmean": ebins_logmean,
        "epd_ebin_labels": ebin_labels,
    }

    return calibration


def get_epd_calibration(probe, instrument, trange):

    if instrument == "epde":
        filename = f"el{probe}_epde_cal_data.txt"

        cal_datasets = None
        with importlib.resources.path("pyspedas.projects.elfin.epd", filename) as cal_file_path:
            cal_datasets = read_epde_calibration_data(cal_file_path)

        if not cal_datasets:
            raise ValueError("Could not load calibration data")

        # Choose the latest calibration data preceeding the start of trange
        cal_datasets.sort(key=lambda cal: cal["date"])
        filtered = list(filter(lambda cal: cal["date"] < time_double(trange[0]), cal_datasets))
        epde_cal_file_data = filtered[-1]

        cal_date = time_string(epde_cal_file_data["date"])
        logging.info(f"Using EPDE calibration data from {cal_date}")

        return get_epde_calibration(epde_cal_file_data)

    elif instrument == "epdi":
        return get_epdi_calibration()
    else:
        raise ValueError(f"Unknown instrument: {instrument!r}")


def calibrate_epd(
    tplotname,
    trange=None,
    type_="eflux",
    probe=None,
    deadtime_corr=None,
    nspinsinsum=1,
):
    """
    Ported from IDL SPEDAS elf_cal_epd procedure. Originally authored
    by Colin Wilkins (colinwilkins@ucla.edu). Relevant comments on
    calibration details have been reproduced here.
    """

    probe = probe if probe is not None else tplotname[2]

    sc = f"el{probe}"
    if "pef" in tplotname:
        instrument = "epde"
        stype = "pef"
    elif "pif" in tplotname:
        instrument = "epdi"
        stype = "pif"

    d = get(tplotname)
    if not d:
        return # TODO: Or raise ValueError?

    trange = time_double(trange) if trange is not None else get_timespan(tplotname)

    dspinper = get_data(f"{sc}_{stype}_spinper")
    dsectnum = get_data(f"{sc}_{stype}_sectnum")
    dnsectors = get_data(f"{sc}_{stype}_nsectors")

    epd_cal = get_epd_calibration(probe=probe, instrument=instrument, trange=trange)

    num_samples = len(d.times)
    cal_ch_factors = epd_cal["epd_cal_ch_factors"]
    overint_factors = epd_cal["epd_overaccumulation_factors"]
    ebins = epd_cal["epd_ebins"]
    ebins_logmean = epd_cal["epd_ebins_logmean"]
    n_sectors = dnsectors.y

    dE = 1.0e-3 * (ebins[1:16] - ebins[0:15]) # in MeV
    dE = np.append(dE, 6.2)

    mytimesra = np.ones(num_samples)
    n_energies = len(ebins)
    mynrgyra = np.ones(n_energies)

    if deadtime_corr is None:
        logging.info("Deadtime correction applied with default deadtime; "
                     "to not apply set deadtime_corr= 0. or a tiny value, e.g. 1.e-9")
        # [default ~ 2% above ~max cps in data (after overaccum. corr.)
        # of 125Kcps corresponds to 8.e-6 us peak hold in front preamp]
        deadtime_corr = 1.0 / (1.02 * 1.25e5)

    dt = (nspinsinsum * dspinper.y / n_sectors * overint_factors[dsectnum.y])
    dt = dt[:,np.newaxis] @ mynrgyra[np.newaxis]

    if type_ == "raw":
        store_data(
            tplotname,
            data={"x": d.times, "y": d.y, "v": np.arange(16, dtype=np.float32)},
            attr_dict=get_data(tplotname, metadata=True)
        )
    elif type_ == "cps":
        cps = d.y / dt
        tot_cps = np.nansum(cps, axis=1)[:,np.newaxis] @ mynrgyra[np.newaxis,:]

        # Deadtime correction
        cps_deadtime_corrected = cps / (1.0 - tot_cps * deadtime_corr)

        # Only reason for negatives is deadtime correction
        ineg = np.where(np.nansum(cps_deadtime_corrected, axis=1) < 0.0)[0]
        if len(ineg) > 0:
            cps_deadtime_corrected[ineg,:] = 0.0
            d3interpol = cps_deadtime_corrected.copy()
            d3interpol[1:num_samples-1,:] = (cps_deadtime_corrected[0:num_samples-2,:] + 
                                             cps_deadtime_corrected[2:num_samples,:]) / 2.0
            cps_deadtime_corrected[ineg,:] = d3interpol[ineg,:]

        store_data(
            tplotname,
            data={"x": d.times, "y": cps_deadtime_corrected, "v": ebins_logmean},
            attr_dict=get_data(tplotname, metadata=True)
        )

    elif type_ == "nflux":
        cps = d.y / dt
        tot_cps = np.nansum(cps, axis=1)[:,np.newaxis] @ mynrgyra[np.newaxis,:]

        # Deadtime correction
        cps_deadtime_corrected = cps / (1.0 - tot_cps * deadtime_corr)

        rel_energy_cal = cal_ch_factors / dE
        nflux = cps_deadtime_corrected * (mytimesra[:,np.newaxis] @ rel_energy_cal[np.newaxis,:])

        # Only reason for negatives is deadtime correction
        ineg = np.where(np.nansum(nflux, axis=1) < 0.0)[0]
        if len(ineg) > 0:
            nflux[ineg,:] = 0.0
            d3interpol = nflux.copy()
            d3interpol[1:num_samples-1,:] = (nflux[0:num_samples-2,:] + 
                                             nflux[2:num_samples,:]) / 2.0
            nflux[ineg,:] = d3interpol[ineg,:]

        store_data(
            tplotname,
            data={"x": d.times, "y": nflux, "v": ebins_logmean},
            attr_dict=get_data(tplotname, metadata=True)
        )

    elif type_ == "eflux":
        cps = d.y / dt
        tot_cps = np.nansum(cps, axis=1)[:,np.newaxis] @ mynrgyra[np.newaxis,:]

        # Deadtime correction
        cps_deadtime_corrected = cps / (1.0 - tot_cps * deadtime_corr)

        eflux_cal = ebins_logmean * (cal_ch_factors / dE)
        eflux = cps_deadtime_corrected * (mytimesra[:,np.newaxis] @ eflux_cal[np.newaxis,:])

        # Only reason for negatives is deadtime correction
        ineg = np.where(np.nansum(eflux, axis=1) < 0.0)[0]
        if len(ineg) > 0:
            eflux[ineg,:] = 0.0
            d3interpol = eflux.copy()
            d3interpol[1:num_samples-1,:] = (eflux[0:num_samples-2,:] + 
                                             eflux[2:num_samples,:]) / 2.0
            eflux[ineg,:] = d3interpol[ineg,:]

        store_data(
            tplotname,
            data={"x": d.times, "y": eflux, "v": ebins_logmean},
            attr_dict=get_data(tplotname, metadata=True)
        )


import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data
import pyspedas


class Dscovr_LoadTestCases(unittest.TestCase):
    def test_load_mag_data(self):
        del_data()
        mag_vars = pyspedas.projects.dscovr.mag(time_clip=True)
        self.assertTrue(data_exists("dsc_h0_mag_B1RTN"))
        self.assertTrue(data_exists("dsc_h0_mag_B1GSE"))

    def test_load_mag_notplot(self):
        del_data()
        mag_vars = pyspedas.projects.dscovr.mag(notplot=True)
        self.assertTrue("dsc_h0_mag_B1GSE" in mag_vars)

    def test_load_fc_data(self):
        del_data()
        fc_vars = pyspedas.projects.dscovr.fc()
        self.assertTrue(data_exists("dsc_h1_fc_Np"))
        self.assertTrue(data_exists("dsc_h1_fc_THERMAL_TEMP"))

    def test_load_orb_data(self):
        del_data()
        orb_vars = pyspedas.projects.dscovr.orb()
        self.assertTrue(data_exists("dsc_orbit_GSE_POS"))
        self.assertTrue(data_exists("dsc_orbit_GCI_POS"))

    def test_load_att_data(self):
        del_data()
        att_vars = pyspedas.projects.dscovr.att()
        self.assertTrue(data_exists("dsc_att_GCI_Yaw"))
        self.assertTrue(data_exists("dsc_att_GCI_Pitch"))
        self.assertTrue(data_exists("dsc_att_GCI_Roll"))

    def test_load_pre_at_data(self):
        del_data()
        pre_at_vars = pyspedas.projects.dscovr.load(
            instrument="pre_at", trange=["2016-01-15", "2016-01-16"], prefix="prevdata_"
        )
        self.assertTrue(data_exists("prevdata_dsc_at_pre_J2000_Yaw"))
        self.assertTrue(data_exists("prevdata_dsc_at_pre_J2000_Pitch"))

    def test_load_downloadonly(self):
        del_data()
        files = pyspedas.projects.dscovr.mag(
            downloadonly=True, trange=["2018-12-15", "2018-12-16"]
        )
        self.assertTrue(os.path.exists(files[0]))

    def test_load_all(self):
        del_data()
        t_all = pyspedas.projects.dscovr.all()
        self.assertTrue(len(t_all) > 0)

    def test_load_invalid_trange(self):
        del_data()
        t_all = pyspedas.projects.dscovr.all(trange="bad input")
        self.assertTrue(len(t_all) == 0)
        del_data()
        t_all = pyspedas.projects.dscovr.all(trange=["2018-12-15"])
        self.assertTrue(len(t_all) == 0)
        del_data()
        t_all = pyspedas.projects.dscovr.all(trange=["2018-12-15", "2018-12-14"])
        self.assertTrue(len(t_all) == 0)

    def test_load_invalid_instrument(self):
        del_data()
        t_all = pyspedas.projects.dscovr.all(instrument="bad input")
        self.assertTrue(len(t_all) == 0)


if __name__ == "__main__":
    unittest.main()



'''
Convert spherical coordinates (r, theta, phi) to Cartesian coordinates (x, y, z).

This function is similar to sphere_to_cart.pro in IDL SPEDAS.
'''
import numpy as np


def sphere_to_cart(r, theta, phi):
    """
    Convert spherical coordinates in degrees (r, theta, phi) to Cartesian coordinates (x, y, z).

    Parameters:
    -----------
    r : float
        Radial distance from the origin.
    theta : float
        Altitude angle (not co-latitude!) It ranges from -90 to 90 degrees
    phi : float
        Azimuthal angle (angle in the xy-plane from the positive x-axis to the point) in degrees. It ranges from [0, 360].

    Returns:
    --------
    x : float
        X coordinate in Cartesian system.
    y : float
        Y coordinate in Cartesian system.
    z : float
        Z coordinate in Cartesian system.

    Notes:
    ------
    - Uses the following equations for conversion:
        x = r*cos(theta)*cos(phi)
        y = r*cos(theta)*sin(phi)
        z = r*sin(theta)
    """
    r, theta, phi = np.array(r), np.deg2rad(theta), np.deg2rad(phi)
    c = np.cos(theta)
    x = r * c * np.cos(phi)
    y = r * c * np.sin(phi)
    z = r * np.sin(theta)

    return x, y, z



import os
import logging
import unittest
from pyspedas.tplot_tools import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        files = pyspedas.projects.wind.mfi(trange=['2013-11-5', '2013-11-6'], downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_3dp_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2003-09-5', '2003-09-6'], notplot=True)
        self.assertTrue('N_e_dens_wi_3dp' in tdp_vars)

    def test_load_3dp_pm_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2018-11-06', '2018-11-07'], datatype='3dp_pm', time_clip=True, notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_pm_P_VELS' in tdp_vars)

    def test_load_3dp_ehpd_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2018-11-06', '2018-11-07'], datatype='3dp_ehpd', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_ehpd_FLUX' in tdp_vars)

    def test_load_3dp_ehsp_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_ehsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_ehsp_FLUX' in tdp_vars)

    def test_load_3dp_elm2_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_elm2', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_elm2_FLUX' in tdp_vars)

    def test_load_3dp_elpd_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_elpd', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_elpd_FLUX' in tdp_vars)

    def test_load_3dp_elsp_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_elsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_elsp_FLUX' in tdp_vars)

    def test_load_3dp_em_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_em', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_em_E_DENS' in tdp_vars)

    def test_load_3dp_emfits_e0_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2003-10-11', '2003-10-12'], datatype='3dp_emfits_e0', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('N_e_dens_wi_3dp' in tdp_vars)

    def test_load_3dp_emfits_e0_data_prefix_suffix(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2003-10-11', '2003-10-12'], datatype='3dp_emfits_e0', time_clip=True,
                                         prefix='pre_', suffix='_suf', notplot=True, addmaster=True)
        self.assertTrue('pre_N_e_dens_wi_3dp_suf' in tdp_vars)

    def test_load_3dp_k0_data(self):
        # Note: this datatype is periodically pruned, SPDF seems to keep the last 3 years worth.
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2023-01-01', '2023-01-02'], datatype='3dp_k0', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_k0_elect_density' in tdp_vars)

    def test_load_3dp_phsp_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_phsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_phsp_FLUX' in tdp_vars)
    def test_varformat_star(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_phsp', time_clip=True,
                                         notplot=True, addmaster=True, varformat='*')
        self.assertTrue('wi_3dp_phsp_FLUX' in tdp_vars)

    def test_varformat_dots(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_phsp', time_clip=True,
                                         notplot=True, addmaster=True, varformat='....')
        self.assertTrue('wi_3dp_phsp_FLUX' in tdp_vars)

    def test_load_3dp_plsp_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_plsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_plsp_FLUX' in tdp_vars)

    def test_load_3dp_varformat_alternation(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_plsp', time_clip=True,
                                         notplot=True, addmaster=True,varformat=r'MOM\.P\.FLUX|MOM\.P\.VELOCITY|MOM\.P\.PTENS')
        self.assertTrue('wi_3dp_plsp_MOM.P.FLUX' in tdp_vars)
        self.assertTrue('wi_3dp_plsp_MOM.P.VELOCITY' in tdp_vars)
        self.assertTrue('wi_3dp_plsp_MOM.P.PTENS' in tdp_vars)
        # Make sure we only got what we asked for
        self.assertFalse('wi_3dp_plsp_FLUX' in tdp_vars)

    def test_load_3dp_sfpd_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sfpd', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sfpd_FLUX' in tdp_vars)

    def test_load_3dp_sfpd_data_prefix_none(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sfpd', time_clip=True,
                                         prefix=None, notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sfpd_FLUX' in tdp_vars)

    def test_load_3dp_sfpd_data_suffix_none(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sfpd', time_clip=True,
                                         suffix=None, notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sfpd_FLUX' in tdp_vars)

    def test_load_3dp_sfpd_data_prefix_suffix(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sfpd', time_clip=True,
                                         prefix='pre_', suffix='_suf', notplot=True, addmaster=True)
        self.assertTrue('pre_wi_3dp_sfpd_FLUX_suf' in tdp_vars)

    def test_load_3dp_sfsp_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sfsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sfsp_FLUX' in tdp_vars)

    def test_load_3dp_sopd_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sopd', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sopd_FLUX' in tdp_vars)

    def test_load_3dp_sosp_data(self):
        tdp_vars = pyspedas.projects.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sosp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sosp_FLUX' in tdp_vars)

    def test_load_mfi_data(self):
        mfi_vars = pyspedas.projects.wind.mfi(trange=['2013-11-5', '2013-11-6'], time_clip=True)
        self.assertTrue('BGSE' in mfi_vars)

    def test_load_swe_data(self):
        swe_vars = pyspedas.projects.wind.swe(trange=['2013-11-5', '2013-11-6'])
        self.assertTrue('N_elec' in swe_vars)
        self.assertTrue('T_elec' in swe_vars)
        self.assertTrue('W_elec' in swe_vars)

    def test_load_waves_data(self):
        wav_vars = pyspedas.projects.wind.waves(trange=['2013-11-5', '2013-11-6'])
        self.assertTrue('E_VOLTAGE_RAD1' in wav_vars)
        self.assertTrue('E_VOLTAGE_RAD2' in wav_vars)
        self.assertTrue('E_VOLTAGE_TNR' in wav_vars)

    def test_load_waves_rad1_rad2_tnr_data(self):
        import pyspedas
        from pyspedas.tplot_tools import tplot, tplot_names
        # example of a Type III radio burst
        trange = ['2019-04-02/15:00', '2019-04-02/16:30']
        rad2_vars = pyspedas.projects.wind.waves(trange=trange, time_clip=True, datatype='rad2')
        rad1_vars = pyspedas.projects.wind.waves(trange=trange, time_clip=True, datatype='rad1')
        tnr_vars = pyspedas.projects.wind.waves(trange=trange, time_clip=True, datatype='tnr')
        self.assertTrue('wi_l2_wav_rad1_PSD_V2_Z' in rad1_vars)
        self.assertTrue(data_exists('wi_l2_wav_rad1_PSD_V2_Z'))
        self.assertTrue('wi_l2_wav_rad2_PSD_V2_Z' in rad2_vars)
        self.assertTrue(data_exists('wi_l2_wav_rad2_PSD_V2_Z'))
        self.assertTrue('wi_l2_wav_tnr_PSD_V2' in tnr_vars)
        self.assertTrue(data_exists('wi_l2_wav_tnr_PSD_V2'))
        #tplot_names()
        #tplot('wi_l2_wav_*_PSD_V2*')

    def test_load_waves_qtn(self):
        trange = ['2005-01-01/00:00', '2005-01-02/00:00']
        qtnfit_vars = pyspedas.projects.wind.waves(trange=trange, time_clip=True, datatype='qtnfit')
        qtnfit_filtered_vars = pyspedas.projects.wind.waves(trange=trange, time_clip=True, datatype='qtnfit-filtered')
        self.assertTrue('wi_l2_wav_qtnfit_fit_n_e' in qtnfit_vars)
        self.assertTrue(data_exists('wi_l2_wav_qtnfit_fit_n_e'))
        self.assertTrue('wi_l2_wav_qtnfit-filtered_fit_n_e' in qtnfit_filtered_vars)
        self.assertTrue(data_exists('wi_l2_wav_qtnfit-filtered_fit_n_e'))


    def test_load_orbit_pre_data(self):
        orb_vars = pyspedas.projects.wind.orbit(trange=['2013-11-5', '2013-11-6'])
        self.assertTrue('GSM_POS' in orb_vars)
        self.assertTrue('GSM_VEL' in orb_vars)
        self.assertTrue('SUN_VECTOR' in orb_vars)
        self.assertTrue('GCI_POS' in orb_vars)
        self.assertTrue('GCI_VEL' in orb_vars)

    def test_load_orbit_def_data(self):
        orb_vars = pyspedas.projects.wind.orbit(trange=['1997-06-5', '1997-06-6'],datatype='def_or')
        self.assertTrue('GSM_POS' in orb_vars)
        self.assertTrue('GSM_VEL' in orb_vars)
        self.assertTrue('SUN_VECTOR' in orb_vars)
        self.assertTrue('GCI_POS' in orb_vars)
        self.assertTrue('GCI_VEL' in orb_vars)

    def test_load_sms_data(self):
        sms_vars = pyspedas.projects.wind.sms()
        self.assertTrue('Alpha_vel' in sms_vars)
        self.assertTrue('C/O_ratio' in sms_vars)
        self.assertTrue('C_ion_temp' in sms_vars)
        self.assertTrue('O_ion_temp' in sms_vars)


if __name__ == '__main__':
    unittest.main()



'''
Convert Solar Magnetic (SM) coordinates to Magnetic Local Time (MLT).

This function is similar to sm2mlt.pro in IDL SPEDAS.
'''
import numpy as np
from pyspedas import cart2spc


def sm2mlt(x_sm, y_sm, z_sm):
    """
    Convert Solar Magnetic (SM) coordinates to Magnetic Local Time (MLT).

    Parameters:
    -----------
    x_sm : float
        X coordinate in Solar Magnetic system.
    y_sm : float
        Y coordinate in Solar Magnetic system.
    z_sm : float
        Z coordinate in Solar Magnetic system.

    Returns:
    --------
    mlt_0 : float
        Magnetic Local Time.

    Notes:
    ------
    - First, the SM coordinates are converted to spherical coordinates.
    - The azimuthal angle from the conversion (phi) is then used to calculate MLT.
    - If the resulting MLT is greater than 24, it is corrected by subtracting 24.
    """

    # Convert to spherical coordinates
    r, theta, phi = cart2spc(x_sm, y_sm, z_sm)

    # Calculate initial MLT
    mlt_0 = 12.0 + phi * 24.0 / (2.0 * np.pi)

    # Fix MLTs greater than 24
    mlt_0[mlt_0 > 24.0] -= 24.0

    return mlt_0



import pyspedas
from pyspedas.tplot_tools import store_data, get_data, degap, del_data
import numpy as np


def makegap(var_data, dt=None, margin=0.0, func="nan"):
    """
    Fill gaps in the data either with NaNs or the last number.  This
    is identical to degap, except operates directly on the data and
    time arrays, rather than the tplot variable. This is intended for
    use with the data_gap option. This version actually puts the data
    into a temporary tplot variable, and call degap, then extracts
    that data into the proper form.

    Parameters
    ----------
        var_data : named tuple
            The data for the tplot variable, a structure that contains at least, tags for 'y' and 'times'
        dt : int/float
            Step size of the data in seconds, default is to use the median time interval
        margin : int/float, optional, default is 0.0 seconds (there is no margin in the IDL tplot makegap)
            The maximum deviation from the step size allowed before degapping occurs.  In other words,
            if you'd like to fill in data every 4 seconds but occasionally the data is 4.1 seconds apart,
            set the margin to .1 so that a data point is not inserted there.
        func : str, optional
            Either 'nan' or 'ffill', which overrides normal interpolation with NaN
            substitution or forward-filled values.

    Returns
    -------
    tuple
        A tuple returned by calling get_data() on the degapped temp variable

    Examples
    --------

        >>> import pyspedas
        >>> time = [pyspedas.time_float("2020-01-01") + i for i in [1, 2, 3, 4, 5, 6, 9, 10, 11]]
        >>> y = [1, 2, 3, 4, 5, 6, 9, 10, 11]
        >>> pyspedas.store_data("a", data={"x": time, "y": y})
        >>> a = pyspedas.get_data("a")
        >>> print(a)
        >>> b = pyspedas.makegap(a)
        >>> print(b)

    """
    # vt = var_data.times
    # gap_size = np.diff(vt) #vt is in nanoseconds, and np.diff works over day boundaries
    # Default for dt is the median value of gap_size, the time interval differences
    # if dt == None:
    #    dt = np.median(gap_size)
    #    dt0 = dt
    # else:
    # change dt to appropriate type
    #    dt0 = np.timedelta64(int(dt*1e9), 'ns')

    # gap_index_locations = np.where(gap_size > dt0)

    # First create a temporary tplot variable for the data, times need to be in unix time
    #x = time_float(var_data.times)
    x = np.int64(var_data.times)/1e9
    if var_data.y.ndim == 1:
        store_data("makegap_tmp", data={"x": x, "y": var_data.y})
    else:  # multiple dimensions
        var_data_out = None
        store_data("makegap_tmp", data={"x": x, "y": var_data.y})
        # Check for values, v, or v1, v2, v3
        if len(var_data) == 2:  # No v's
            store_data("makegap_tmp", data={"x": x, "y": var_data.y})
        elif len(var_data) == 3:  # v or v1,needs indexing if it's 2d
            store_data(
                "makegap_tmp", data={"x": x, "y": var_data.y, "v": var_data[2]}
            )
        elif len(var_data) == 4:  # has both v1 and v2
            store_data(
                "makegap_tmp",
                data={"x": x, "y": var_data.y, "v1": var_data[2], "v2": var_data[3]},
            )
        elif len(var_data) == 5:  # has v1, v2, v3
            store_data(
                "makegap_tmp",
                data={
                    "x": x,
                    "y": var_data.y,
                    "v1": var_data[2],
                    "v2": var_data[3],
                    "v3": var_data[3],
                },
            )
    # Now, degap the variable
    degap("makegap_tmp", dt=dt, margin=margin, func=func)
    # and return the getdata result
    var_data_out = get_data("makegap_tmp", dt=True)
    del_data("makegap_tmp")
    return var_data_out




from .spacecraft.fields.fgm import fgm
from .spacecraft.fields.fit import fit
from .spacecraft.fields.efi import efi
from .spacecraft.fields.fft import fft
from .spacecraft.fields.fbk import fbk
from .spacecraft.fields.scm import scm

from .spacecraft.particles.esa import esa
from .spacecraft.particles.esd import esd
from .spacecraft.particles.sst import sst
from .spacecraft.particles.mom import mom
from .spacecraft.particles.gmom import gmom

from .ground.gmag import gmag
from .ground.ask import ask

from .state_tools.state import state
from .state_tools.slp import slp
from .state_tools.ssc import ssc
from .state_tools.ssc_pre import ssc_pre

from .state_tools.autoload_support import autoload_support
from .state_tools import get_spinmodel
from .cotrans import sse2sel,gse2sse,dsl2gse,ssl2dsl

from .analysis import scpot2dens, scpot2dens_nishimura



import os

CONFIG = {'local_data_dir': 'akebono_data/',
          'remote_data_dir': 'https://data.darts.isas.jaxa.jp/pub/akebono/'}

# override local data directory with environment variables
if os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir'] = os.sep.join([os.environ['SPEDAS_DATA_DIR'], 'akebono'])

if os.environ.get('AKEBONO_DATA_DIR'):
    CONFIG['local_data_dir'] = os.environ['AKEBONO_DATA_DIR']
    


from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG

def load(trange=['2013-11-5', '2013-11-6'], 
         instrument='mgf',
         datatype='k0', 
         suffix='', 
         prefix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         force_download=False):
    """
    This function loads data from the Geotail mission; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.projects.geotail.mgf
        pyspedas.projects.geotail.efd
        pyspedas.projects.geotail.lep
        pyspedas.projects.geotail.cpi
        pyspedas.projects.geotail.epi
        pyspedas.projects.geotail.pwi

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2013-11-5', '2013-11-6']

        instrument: str
            Spacecraft identifier ('mgf', 'efd', 'lep', 'cpi', 'edi', 'pwi')
            Default: 'mgf'

        datatype: str
            Data type; Valid options: ('k0', ''eda3sec', 'edb3sec')
            Default: 'k0'

        suffix: str
            The tplot variable names will be given this suffix.
            Default: no suffix is added

        prefix: str
            The tplot variable names will be given this prefix.
            Default: no prefix is added

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: all variables are loaded in

        varnames: list of str
            List of variable names to load
            Default: all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    ----------
    list of str
        List of tplot variables created.

    Example
    ----------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> mgf_vars = pyspedas.projects.geotail.mgf(trange=['2018-11-5', '2018-11-6'])
    >>> tplot(['IB', 'IB_vector'])

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> efd_vars = pyspedas.projects.geotail.efd(trange=['2018-11-5', '2018-11-6'])
    >>> tplot(['Es', 'Ss', 'Bs', 'Vs', 'Ew', 'Sw', 'Bw', 'Vw'])

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> lep_vars = pyspedas.projects.geotail.lep(trange=['2018-11-5/05:00', '2018-11-5/06:00'], time_clip=True)
    >>> tplot(['N0', 'V0'])

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> cpi_vars = pyspedas.projects.geotail.cpi(trange=['2018-11-5/15:00', '2018-11-5/18:00'], time_clip=True)
    >>> tplot(['SW_P_Den', 'SW_P_AVGE', 'SW_V', 'HP_P_Den'])

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> epic_vars = pyspedas.projects.geotail.epic(trange=['2018-11-5', '2018-11-6'])
    >>> tplot('IDiffI_I')

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> pwi_vars = pyspedas.projects.geotail.pwi(trange=['2018-11-5/06:00', '2018-11-5/07:00'], time_clip=True)
    >>> tplot(['MCAE_AVE', 'MCAB_AVE'])

    """

    if instrument == 'mgf':
        if datatype == 'k0':
            pathformat = 'mgf/mgf_k0/%Y/ge_'+datatype+'_mgf_%Y%m%d_v??.cdf'
        elif datatype == 'eda3sec' or datatype == 'edb3sec':
            pathformat = 'mgf/'+datatype+'_mgf/%Y/ge_'+datatype+'_mgf_%Y%m%d_v??.cdf'
    elif instrument == 'efd':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/ge_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'lep':
        if datatype == 'k0':
            pathformat = 'lep/lep_k0/%Y/ge_'+datatype+'_lep_%Y%m%d_v??.cdf'
    elif instrument == 'cpi':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/ge_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'epi':
        pathformat = 'epic/'+instrument+'_'+datatype+'/%Y/ge_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'pwi':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/ge_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, prefix=prefix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import logging
import warnings
from pyspedas.tplot_tools import get_data, store_data, options, time_double
import numpy as np
from hapiclient import hapi as load_hapi
from .replace_fillvals import replace_fillvals


def hapi(trange=None, server=None, dataset=None, parameters='', suffix='',
         prefix='', catalog=False, quiet=False):
    """
    Loads data from a HAPI server into tplot variables

    Parameters
    -----------
        trange: list of str or list of float
            Time range to load the data for

        server: str
            HAPI server to load the data from

        dataset: str
            HAPI dataset to load

        parameters: str or list of str
            Parameters in the dataset to load; default
            is to load them all
            Default: '' (load all parameters)

        prefix: str
            Prefix to append to the tplot variables
            Default: ''

        suffix: str
            Suffix to append to the tplot variables
            Default: ''

        catalog: bool
            If True, returns the server's catalog of datasets
            Default: False

        quiet: bool
            If True, suppress printing the catalog ids retrieved
            Default: False

    Returns
    -------
        list of str
            List of catalog ids retrieved or tplot variables created.

    Examples
    --------

    >>> # Print catalog from CDAWeb HAPI server
    >>> import pyspedas
    >>> cat_entries = pyspedas.hapi(server='https://cdaweb.gsfc.nasa.gov/hapi', catalog=True)

    >>> # Load OMNI data from CDAWeb HAPI server
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> h_vars = pyspedas.hapi(trange=['2003-10-20', '2003-11-30'],server='https://cdaweb.gsfc.nasa.gov/hapi',dataset='OMNI_HRO2_1MIN')
    >>> tplot(['BX_GSE','BY_GSE','BZ_GSE'])
    """

    if server is None:
        logging.error('No server specified; example servers include:')
        logging.error('- https://cdaweb.gsfc.nasa.gov/hapi')
        logging.error('- https://pds-ppi.igpp.ucla.edu/hapi')
        logging.error('- http://planet.physics.uiowa.edu/das/das2Server/hapi')
        logging.error('- https://iswa.gsfc.nasa.gov/IswaSystemWebApp/hapi')
        logging.error('- http://lasp.colorado.edu/lisird/hapi')
        return

    if catalog:
        catalog = load_hapi(server)
        items = []
        id_list = []
        if 'catalog' in catalog.keys():
            items = catalog['catalog']
        if not quiet:
            print('Available datasets: ')
        for item in items:
            if 'title' in item.keys():
                if not quiet:
                    print(item['id'] + ': ' + item['title'])
            else:
                if not quiet:
                    print(item['id'])
            id_list.append(item['id'])
        return id_list

    if dataset is None:
        logging.error('Error, no dataset specified; please see the catalog for a list of available data sets.')
        return

    if trange is None:
        logging.error('Error, no trange specified')
        return

    if isinstance(parameters, list):
        parameters = ','.join(parameters)

    opts = {'logging': False}

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=ResourceWarning)
        warnings.filterwarnings('ignore', message='Unverified HTTPS request')
        data, hapi_metadata = load_hapi(server, dataset, parameters, trange[0], trange[1], **opts)

    out_vars = []

    # loop through the parameters in this dataset
    params = hapi_metadata['parameters']

    timestamps = [datapoint[0] for datapoint in data]
    unixtimes = [time_double(timestamp.decode('utf-8')) for timestamp in timestamps]

    for param_idx, param in enumerate(params[1:]):
        spec = False
        param_name = param.get('name')
        param_type = param.get('type')
        data_size = param.get('size')

        if param_type is None:
            param_type = 'double'

        if data_size is None:
            single_line = True

        try:
            if param_type == 'double':
                single_line = isinstance(data[0][param_idx+1], np.float64)
            elif param_type == 'integer':
                single_line = isinstance(data[0][param_idx+1], np.int32)
        except IndexError:
            continue

        if single_line:
            data_out = np.zeros((len(data)))
        else:
            try:
                data_out = np.zeros((len(data), len(data[0][param_idx+1])))
            except TypeError:
                continue

        for idx, datapoint in enumerate(data):
            if param_type in ['double','integer']:
                datapt = datapoint[param_idx+1]
            else:
                datapt = None
            if single_line:
                data_out[idx] = datapt
            else:
                data_out[idx, :] = datapt

        data_out = data_out.squeeze()

        # check for fill values
        fill_value = param.get('fill')
        if fill_value is not None:
            replace_fillvals(data_out, fill_value, param_name, param_type)

        bins = param.get('bins')

        if bins is not None:
            centers = bins[0].get('centers')

            if centers is not None:
                spec = True

        data_table = {'x': unixtimes, 'y': data_out}

        if spec:
            data_table['v'] = centers

        saved = store_data(prefix + param_name + suffix, data=data_table)
        metadata = get_data(prefix + param_name + suffix, metadata=True)
        metadata['HAPI'] = param

        if spec:
            options(prefix + param_name + suffix, 'spec', True)

        param_units = param.get('units')
        if param_units is not None:
            options(prefix + param_name + suffix, 'ysubtitle', '[' + str(param_units) + ']')

        param_desc = param.get('description')
        if param_desc is not None:
            options(prefix + param_name + suffix, 'ytitle', param_desc)

        if saved:
            out_vars.append(prefix + param_name + suffix)

    return out_vars



import logging
import numpy as np
from pyspedas.tplot_tools import tnames, get_data, store_data, options
from pyspedas.tplot_tools import pwrspc


def tpwrspc(varname, newname=None, overwrite=False, noline=False, nohanning=False, bin=3, notperhz=False, trange=None, axis=0):
    """
    This function is a wrapper for pwrspc.
    It applies pwrspc to a tplot variable and stores the result in a new tplot variable.

    Parameters
    ----------
        varname: str
            Name of the tplot variable.

        newname: str, optional
            New name for the output tplot variable. If not set, default output variable will be
            Default: None

        overwrite: bool, optional
            If True, overwrite the existing tplot variable.
            If True, then the newname keyword has no effect.
            Default: False

        noline: bool, optional
            If True, straight line is not subtracted from the data.
            Default: False

        nohanning: bool, optional
            If True, no Hanning window is applied to the data.
            Default: False
        
        bin: int, optional
            Bin size for binning the data. 
            Default: 3

        notperhz: bool, optional
            If True, the output units are the square of the input units.
            Default: False

        trange: list, optional
            Time range for the data extraction.
            Default: None

        axis: int, optional
            If the input variable is multi-dimensional, this specifies the axis to operate along.
            Default: 0, first axis.

    Returns
    --------
    str
            Name of the new tplot variable created by this function.
            The output variable contains a single data point, frequency as v, and power as y.
    
    Examples
    --------
        >>> import pyspedas
        >>> import numpy as np
        >>> pyspedas.store_data('a', data={'x': range(100), 'y': np.random.random(100)})
        >>> pyspedas.tpwrspc('a')
        >>> pyspedas.get_data('a_pwrspc')
    """

    # Check if the variable exists
    if tnames(varname) == []:
        logging.info('This tplot variable does not exist.')
        return ''

    # Check for conflicting arguments
    if overwrite:
        logging.info('Variable will be overwritten.')
        newname = varname
    else:
        if newname:
            if tnames(newname) != []:
                logging.info('This new tplot variable already exists. Please use overwrite=True to overwrite it.')
                return ''
        else:
            newname = varname + '_pwrspc'

    # Get data from tplot variable
    d = get_data(varname)
    t = np.array(d[0], dtype='float64')
    data = np.array(d[1], dtype='float64')

    # If the input variable is multi-dimensional, operate along the specified axis
    if data.ndim == 1:
        y = data
    elif data.ndim == 2:
        if axis < 1:
            axis = 0
        elif axis > data.shape[1]:
            axis = data.shape[1] - 1
        logging.info(f'Operating along axis {axis}')
        y = data[:, axis]
    else:
        logging.info('Cannot handle data with more than two dimensions.')
        return ''

    # Remove NaN points
    tav = np.mean(t)  # We are going to store f,p at the time average point
    goodpoints = ~np.isnan(t) & ~np.isnan(y)
    if np.any(goodpoints):
        t = t[goodpoints]
        y = y[goodpoints]
        logging.info('NaN points have been removed.')
    else:
        logging.info(f'No valid data points in {varname}')
        return ''

    # Restrict to a time range if specified
    if trange and len(trange) == 2:
        ok = (trange[0] <= t < trange[1])
        if not np.any(ok):
            logging.info('No data in time range for:', varname)
            logging.info('No Power spectrum for:', varname)
            return ''
        else:
            t = t[ok]
            y = y[ok]

    # Call the power spectrum function (previously defined)
    t = t-t[0]
    f, p = pwrspc(t, y, noline=noline, nohanning=nohanning, bin=bin, notperhz=notperhz)

    if f is not None and p is not None and len(f) > 0 and len(p) > 0 and len(f) == len(p):
        # Store the result in a new tplot variable
        # A single data point will be created, with frequency as v, and power as y
        pp = np.array([p,])
        ff = np.array([f,])
        tt = np.array([tav,], dtype='float64')
        store_data(newname, data={'x': tt, 'y':  pp, 'v':  ff})
        options(newname, 'spec', 1)
        options(newname, 'ylog', 1)
    else:
        print('No Power spectrum for:', varname)
        return ''

    return newname



import pyspedas
import numpy as np
from pyspedas.tplot_tools import get_y_range
import logging


def replace_data(tplot_name, new_data):
    """
    This function will replace all of the data in a tplot variable.

    Parameters
    ----------
    tplot_name : str
        The name of the tplot variable.
    new_data : array_like
        The new data to replace the existing data.
        This can be any object that can be converted into an np.array.

    Returns
    -------
    None

    Examples
    --------
    >>> # Replace data into an existing variable
    >>> import pyspedas
    >>> pyspedas.store_data("v1", data={'x':[1,2,3,4],'y':[1,2,3,4]})
    >>> print(pyspedas.get_data('v1'))
    >>> pyspedas.replace_data("v1",[5,6,7,8])
    >>> print(pyspedas.get_data('v1'))
    """

    # if old name input is a number, convert to corresponding name
    if isinstance(tplot_name, int):
        tplot_name = pyspedas.tplot_tools.data_quants[tplot_name].name

    # check if old name is in current dictionary
    if tplot_name not in pyspedas.tplot_tools.data_quants.keys():
        logging.info(f"{tplot_name} is currently not in pyspedas")
        return

    new_data_np = np.asarray(new_data)
    shape_old = pyspedas.tplot_tools.data_quants[tplot_name].values.shape
    shape_new = new_data_np.shape
    if shape_old != shape_new:
        logging.info(
            f"Dimensions do not match for replace data. {shape_new} does not equal {shape_old}.  Returning..."
        )
        return

    pyspedas.tplot_tools.data_quants[tplot_name].values = new_data_np

    pyspedas.tplot_tools.data_quants[tplot_name].attrs["plot_options"]["yaxis_opt"]["y_range"] = (
        get_y_range(pyspedas.tplot_tools.data_quants[tplot_name])
    )

    return



"""Automated tests for tplot math and utility functions."""

import unittest
import math
import numpy as np
from numpy.testing import assert_array_equal
from pyspedas import (
    clean_spikes,
    smooth,
    tsmooth,
    tdotp,
    tcrossp,
    tnormalize,
    subtract_median,
    subtract_average,
    get_data,
    store_data,
    del_data,
    crop,
    add_across,
    add,
    subtract,
    multiply,
    divide,
    derive,
    tres,
    tpwrspc,
    pwr_spec,
    spec_mult,
    data_exists,
    time_float,
)


class BaseTestCase(unittest.TestCase):
    """Data to be used in tests."""

    def setUp(self):
        """Create a tplot variable to be used in tests."""
        store_data(
            "test",
            data={
                "x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                "y": [3.0, 5.0, 8.0, 15.0, 20.0, 1.0],
            },
        )


class AnalysisTestCases(BaseTestCase):
    """Test functions under tplot_math, and some other utilities."""

    def test_tdotp(self):
        store_data("var1", data={"x": [0], "y": [[3, -3, 1]]})
        store_data("var2", data={"x": [0], "y": [[4, 9, 2]]})
        dp = tdotp("var1", "var2")
        dpdata = get_data("var1_dot_var2")
        self.assertTrue(dpdata.y == np.array([-13]))
        dp = tdotp("var1_doesnt_exist", "var2")
        self.assertIsNone(dp)
        dp = tdotp("var1", "var2_doesnt_exist")
        self.assertIsNone(dp)

    def test_tcrossp(self):
        """cross product tests"""
        cp = tcrossp([3, -3, 1], [4, 9, 2], return_data=True)
        self.assertTrue(cp.tolist() == [-15, -2, 39])
        cp = tcrossp([3, -3, 1], [4, 9, 2])
        cp = get_data(cp)
        self.assertTrue(cp.y[0, :].tolist() == [-15, -2, 39])
        store_data("var1", data={"x": [0], "y": [[3, -3, 1]]})
        store_data("var2", data={"x": [0], "y": [[4, 9, 2]]})
        cp = tcrossp("var1", "var2", return_data=True)
        self.assertTrue(cp[0].tolist() == [-15, -2, 39])
        cp = tcrossp("var1", "var2", newname="test_crossp")
        cp = get_data("test_crossp")
        self.assertTrue(cp.y[0, :].tolist() == [-15, -2, 39])

    def test_tnormalize(self):
        """tests for normalizing tplot variables"""
        store_data(
            "test_tnormalize",
            data={
                "x": [1, 2, 3, 4, 5],
                "y": [[3, 2, 1], [1, 2, 3], [10, 5, 1], [8, 10, 14], [70, 20, 10]],
            },
        )
        norm = tnormalize("test_tnormalize")
        normalized_data = get_data(norm)
        self.assertTrue(np.round(normalized_data.y[0, :], 4).tolist() == [0.8018, 0.5345, 0.2673])
        self.assertTrue(np.round(normalized_data.y[1, :], 4).tolist() == [0.2673, 0.5345, 0.8018])
        self.assertTrue(np.round(normalized_data.y[2, :], 4).tolist() == [0.8909, 0.4454, 0.0891])
        self.assertTrue(np.round(normalized_data.y[3, :], 4).tolist() == [0.4216, 0.527, 0.7379])
        self.assertTrue(np.round(normalized_data.y[4, :], 4).tolist() == [0.9526, 0.2722, 0.1361])

    def test_subtract_median(self):
        """Test subtract_median."""
        subtract_median("aaabbbccc")  # Test non-existent name
        subtract_median("test")
        d = get_data("test-m")
        self.assertTrue(d[1].tolist() == [-3.5, -1.5, 1.5, 8.5, 13.5, -5.5])
        dn = [
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [23.0, 15.0, 28.0],
            [15.0, 20.0, float("nan")],
        ]
        store_data("test1", data={"x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0], "y": dn})
        subtract_median("aaabbbcc")
        subtract_median("test1", newname="aabb")
        d = get_data("aabb")
        self.assertTrue(len(d[1]) == 6)
        subtract_median(["test", "aabb"], newname="aaabbb")
        subtract_median("test1", overwrite=1)
        subtract_average("test", newname="testtest")
        subtract_average(["test-m", "test"], newname="testtest2")

    def test_subtract_average(self):
        """Test subtract_average."""
        subtract_average("aaabbbccc")  # Test non-existent name
        subtract_average("test")
        d = get_data("test-d")
        self.assertTrue((np.round(d[1].tolist()) == [-6.0, -4.0, -1.0, 6.0, 11.0, -8.0]).all())
        dn = [
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [23.0, 15.0, 28.0],
            [15.0, 20.0, float("nan")],
        ]
        store_data("test1", data={"x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0], "y": dn})
        subtract_average("aaabbbcc")
        subtract_average("test1", newname="aabb")
        d = get_data("aabb")
        self.assertTrue(len(d[1]) == 6)
        subtract_average(["test", "aabb"], newname="aaabbb")
        subtract_average("test1", overwrite=1)
        subtract_average("test1", newname="testtest")
        subtract_average(["test1", "test"], newname="testtest2")

    def test_clean_spikes(self):
        """Test clean_spikes."""
        clean_spikes("aaabbbccc", nsmooth=3)  # Test non-existent name
        clean_spikes("test", nsmooth=3)
        d = get_data("test-despike")
        self.assertTrue(len(d[1]) == 6)
        # Now test 3 dim data.
        dn = [
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [23.0, 15.0, 28.0],
            [15.0, 20.0, 1.0],
        ]
        store_data("test1", data={"x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0], "y": dn})
        clean_spikes("test1", nsmooth=3)
        d2 = get_data("test1-despike")
        clean_spikes("test", newname="test_desp", nsmooth=3, sub_avg=True)
        clean_spikes(["test", "test1"], newname="test1-desp")
        clean_spikes("test1", overwrite=1)
        self.assertTrue(len(d2[1]) == 6)

    def test_tsmooth(self):
        """Test smooth."""
        tsmooth("aaabbbccc")  # Test non-existent name
        a = [1.0, 1.0, 2.0, 3.0, 4.0, 1.0, 4.0, 3.0, 2.0, 1.0, 1.0]
        x = smooth(a, 3)
        r = [
            1.0,
            1.3333333333333333,
            2.0,
            3.0,
            2.6666666666666665,
            3.0,
            2.6666666666666665,
            3.0,
            2.0,
            1.3333333333333333,
            1.0,
        ]
        self.assertTrue(x == r)
        b = [1.0, 1.0, 2.0, 3.0, np.nan, np.nan, np.nan, np.nan, 2.0, 1.0, 1.0]
        y = smooth(b, width=3)
        ry = [
            1.0,
            1.3333333333333333,
            2.0,
            1.6666666666666665,
            1.0,
            np.nan,
            np.nan,
            0.6666666666666666,
            1.0,
            1.3333333333333333,
            1.0,
        ]
        self.assertTrue(y == ry)
        tsmooth("test")
        d = get_data("test-s")
        tsmooth("test", overwrite=1)
        tsmooth("test", newname="testtest")
        tsmooth(["test", "test-s"], newname="testtest2")
        self.assertTrue(d[1].tolist() == [3.0, 5.0, 8.0, 15.0, 20.0, 1.0])

    def test_tplot_arithmetic(self):
        del_data("*")
        times = [0.0, 1.0, 2.0, 3.0, 4.0]
        dat1 = [1.0, 2.0, 3.0, 4.0, 5.0]
        dat2 = [10.0, 20.0, 30.0, 40.0, 50.0]
        store_data("v1", data={"x": times, "y": dat1})
        store_data("v2", data={"x": times, "y": dat2})
        add("v1", "v2", newname="sum")
        subtract("v2", "v1", newname="diff")
        multiply("v1", "v2", newname="prod")
        divide("v2", "v1", newname="quot")
        derive("v2", newname="ddx")
        sum_t, sum_d = get_data("sum")
        diff_t, diff_d = get_data("diff")
        prod_t, prod_d = get_data("prod")
        quot_t, quot_d = get_data("quot")
        ddx_t, ddx_d = get_data("ddx")
        assert_array_equal(diff_d, np.array([9.0, 18.0, 27.0, 36.0, 45.0]))
        assert_array_equal(sum_d, np.array([11.0, 22.0, 33.0, 44.0, 55.0]))
        assert_array_equal(quot_d, np.array([10.0, 10.0, 10.0, 10.0, 10.0]))
        assert_array_equal(prod_d, np.array([10.0, 40.0, 90.0, 160.0, 250.0]))
        assert_array_equal(ddx_d, np.array([10.0, 10.0, 10.0, 10.0, 10.0]))
        dt = tres("v1")
        self.assertEqual(dt, 1.0)

    def test_tplot_spectools(self):
        del_data("*")
        # tpwrspc, pwrspc
        store_data("a", data={"x": range(100), "y": np.random.random(100)})
        tpwrspc("a")
        spec = get_data("a_pwrspc")
        self.assertEqual(len(spec.y), 1)
        self.assertTrue(data_exists("a_pwrspc"))
        # pwr_spec
        time = [time_float("2020-01-01") + i for i in range(10000)]
        quantity = [math.sin(i) for i in range(10000)]
        store_data("dp", data={"x": time, "y": quantity})
        pwr_spec("dp", newname="dp_pwrspec")
        # pyspedas.tplot("dp_pwrspec")
        self.assertTrue(data_exists("dp_pwrspec"))
        # spec_mult
        store_data(
            "h",
            data={
                "x": [0, 4, 8, 12, 16, 19, 21],
                "y": [
                    [8, 1, 1],
                    [100, 2, 3],
                    [4, 2, 47],
                    [4, 39, 5],
                    [5, 5, 99],
                    [6, 6, 25],
                    [7, -2, -5],
                ],
                "v": [
                    [1, 1, 50],
                    [2, 2, 3],
                    [100, 4, 47],
                    [4, 90, 5],
                    [5, 5, 99],
                    [6, 6, 25],
                    [7, 7, -5],
                ],
            },
        )
        spec_mult("h")
        hdat = get_data("h_specmult")
        assert_array_equal(
            hdat.y,
            np.array(
                [
                    [8, 1, 50],
                    [200, 4, 9],
                    [400, 8, 2209],
                    [16, 3510, 25],
                    [25, 25, 9801],
                    [36, 36, 625],
                    [49, -14, 25],
                ]
            ),
        )

    def test_add_across(self):
        del_data("*")
        # Add across every column in the data
        store_data(
            "d",
            data={
                "x": [2, 5, 8, 11, 14, 17, 21],
                "y": [
                    [1, 1, 50],
                    [2, 2, 3],
                    [100, 4, 47],
                    [4, 90, 5],
                    [5, 5, 99],
                    [6, 6, 25],
                    [7, 7, -5],
                ],
            },
        )
        add_across("d", newname="d_aa")
        daa = get_data("d_aa")
        assert_array_equal(daa.y, np.array([[52], [7], [151], [99], [109], [37], [9]]))

        # Add across specific columns in the data
        store_data(
            "b",
            data={
                "x": [2, 5, 8, 11, 14, 17, 20],
                "y": [
                    [1, 1, 1, 1, 1, 1],
                    [2, 2, 5, 4, 1, 1],
                    [100, 100, 3, 50, 1, 1],
                    [4, 4, 8, 58, 1, 1],
                    [5, 5, 9, 21, 1, 1],
                    [6, 6, 2, 2, 1, 1],
                    [7, 7, 1, 6, 1, 1],
                ],
            },
        )
        add_across("b", column_range=[[1, 2], [3, 4]], newname="b_aap")
        dbaap = get_data("b_aap")
        assert_array_equal(
            dbaap.y,
            np.array([[2, 2], [7, 5], [103, 51], [12, 59], [14, 22], [8, 3], [8, 7]]),
        )

    def test_crop(self):
        del_data("*")
        store_data("a", data={"x": [0, 4, 8, 12, 16], "y": [1, 2, 3, 4, 5]})
        store_data(
            "b",
            data={
                "x": [2, 5, 8, 11, 14, 17, 20],
                "y": [
                    [1, 1, 1, 1, 1, 1],
                    [2, 2, 5, 4, 1, 1],
                    [100, 100, 3, 50, 1, 1],
                    [4, 4, 8, 58, 1, 1],
                    [5, 5, 9, 21, 1, 1],
                    [6, 6, 2, 2, 1, 1],
                    [7, 7, 1, 6, 1, 1],
                ],
            },
        )
        crop("a", "b", replace=False)
        acrop_data = get_data("a_cropped")
        bcrop_data = get_data("b_cropped")
        assert_array_equal(acrop_data.times, np.array([4.0, 8.0, 12.0, 16.0]))
        assert_array_equal(acrop_data.y, np.array([2, 3, 4, 5]))
        assert_array_equal(bcrop_data.times, np.array([2.0, 5.0, 8.0, 11.0, 14.0]))
        assert_array_equal(
            bcrop_data.y,
            np.array(
                [
                    [1, 1, 1, 1, 1, 1],
                    [2, 2, 5, 4, 1, 1],
                    [100, 100, 3, 50, 1, 1],
                    [4, 4, 8, 58, 1, 1],
                    [5, 5, 9, 21, 1, 1],
                ]
            ),
        )
        # with replacement
        crop("a", "b", replace=True)
        acrop_data = get_data("a")
        bcrop_data = get_data("b")
        assert_array_equal(acrop_data.times, np.array([4.0, 8.0, 12.0, 16.0]))
        assert_array_equal(acrop_data.y, np.array([2, 3, 4, 5]))
        assert_array_equal(bcrop_data.times, np.array([2.0, 5.0, 8.0, 11.0, 14.0]))
        assert_array_equal(
            bcrop_data.y,
            np.array(
                [
                    [1, 1, 1, 1, 1, 1],
                    [2, 2, 5, 4, 1, 1],
                    [100, 100, 3, 50, 1, 1],
                    [4, 4, 8, 58, 1, 1],
                    [5, 5, 9, 21, 1, 1],
                ]
            ),
        )


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from pyspedas.tplot_tools import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_celias_data(self):
        out_vars = pyspedas.projects.soho.celias(time_clip=True)
        self.assertTrue(data_exists('V_p'))
        self.assertTrue(data_exists('N_p'))
    
    def test_load_celias_data_prefix_none(self):
        out_vars = pyspedas.projects.soho.celias(time_clip=True, prefix=None)
        self.assertTrue(data_exists('V_p'))
        self.assertTrue(data_exists('N_p'))
    
    def test_load_celias_data_suffix_none(self):
        out_vars = pyspedas.projects.soho.celias(time_clip=True, suffix=None)
        self.assertTrue(data_exists('V_p'))
        self.assertTrue(data_exists('N_p'))

    def test_load_celias_data_prefix_suffix(self):
        out_vars = pyspedas.projects.soho.celias(time_clip=True, prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_V_p_suf'))
        self.assertTrue(data_exists('pre_N_p_suf'))

    def test_load_cosp_data(self):
        out_vars = pyspedas.projects.soho.costep(time_clip=True)
        self.assertTrue(data_exists('P_int'))
        self.assertTrue(data_exists('He_int'))

    def test_load_cosp_data_prefix_none(self):
        out_vars = pyspedas.projects.soho.costep(time_clip=True, prefix=None)
        self.assertTrue(data_exists('P_int'))
        self.assertTrue(data_exists('He_int'))

    def test_load_cosp_data_suffix_none(self):
        out_vars = pyspedas.projects.soho.costep(time_clip=True, suffix=None)
        self.assertTrue(data_exists('P_int'))
        self.assertTrue(data_exists('He_int'))

    def test_load_cosp_data_prefix_suffix(self):
        out_vars = pyspedas.projects.soho.costep(time_clip=True, prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_P_int_suf'))
        self.assertTrue(data_exists('pre_He_int_suf'))    

    def test_load_erne_data(self):
        out_vars = pyspedas.projects.soho.erne(time_clip=True)
        self.assertTrue(data_exists('PH'))

    def test_load_erne_data_prefix_none(self):
        out_vars = pyspedas.projects.soho.erne(time_clip=True, prefix=None)
        self.assertTrue(data_exists('PH'))

    def test_load_erne_data_suffix_none(self):
        out_vars = pyspedas.projects.soho.erne(time_clip=True, suffix=None)
        self.assertTrue(data_exists('PH'))

    def test_load_erne_data_prefix_suffix(self):
        out_vars = pyspedas.projects.soho.erne(time_clip=True, prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_PH_suf'))

    def test_load_orbit_data(self):
        out_vars = pyspedas.projects.soho.orbit(time_clip=True)
        self.assertTrue(data_exists('GSE_POS'))
        self.assertTrue(data_exists('GSE_VEL'))

    def test_load_orbit_data_prefix_none(self):
        out_vars = pyspedas.projects.soho.orbit(time_clip=True, prefix=None)
        self.assertTrue(data_exists('GSE_POS'))
        self.assertTrue(data_exists('GSE_VEL'))

    def test_load_orbit_data_suffix_none(self):
        out_vars = pyspedas.projects.soho.orbit(time_clip=True, suffix=None)
        self.assertTrue(data_exists('GSE_POS'))
        self.assertTrue(data_exists('GSE_VEL'))

    def test_load_orbit_data_prefix_suffix(self):
        out_vars = pyspedas.projects.soho.orbit(time_clip=True, prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_GSE_POS_suf'))
        self.assertTrue(data_exists('pre_GSE_VEL_suf'))

    def test_load_notplot(self):
        out_vars = pyspedas.projects.soho.erne(notplot=True)
        self.assertTrue('PH' in out_vars)

    def test_downloadonly(self):
        files = pyspedas.projects.soho.erne(downloadonly=True, trange=['2006-06-01', '2006-06-02'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()

    


from .load import load
from .orb_postprocessing import orb_postprocessing


# This routine was originally in akebono/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

def orb(trange=['2012-10-01', '2012-10-02'],
        prefix='',
        suffix='',
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        force_download=False):
    """
    Loads Akebono orbit data (orb)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2012-10-01', '2012-10-02']

        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
            Default: None

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)
            Default: []

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    ----------
    list of str
        List of tplot variables created.

    Example
    ----------
        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> orb_vars = pyspedas.projects.akebono.orb(trange=['2012-10-01', '2012-10-02'])
        >>> tplot(['akb_orb_geo', 'akb_orb_MLT'])

    """
    files = load(instrument='orb', trange=trange, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, force_download=force_download)

    if files is None or notplot or downloadonly:
        return files

    return orb_postprocessing(files, prefix=prefix, suffix=suffix)





import unittest
import numpy as np
from numpy.testing import assert_array_equal
from pyspedas import hapi, data_exists, del_data
from pyspedas.hapi_tools.replace_fillvals import replace_fillvals


class HAPITests(unittest.TestCase):
    def test_replace_fillvals(self):
        da1_dbl = np.array([1.0, 2.0, 3.0, 4.0])
        fv1_dbl = 3.0
        replace_fillvals(da1_dbl, fv1_dbl, "da1_dbl_scalarfill", "double")
        assert_array_equal(da1_dbl, [1.0, 2.0, np.nan, 4.0])
        da2_dbl = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 2.0]])
        fv2_dbl = [1.0, 2.0]
        replace_fillvals(da2_dbl, fv2_dbl, "da2_dbl_arrayfill", "double")
        assert_array_equal(
            da2_dbl, np.array([[np.nan, np.nan], [3.0, 4.0], [5.0, 6.0], [7.0, np.nan]])
        )
        # Repeat after replacement, should be nothing to replace, and output array same as before
        fv2_dbl = [1.0, 2.0]
        replace_fillvals(da2_dbl, fv2_dbl, "da2_dbl_arrayfill", "double")
        assert_array_equal(
            da2_dbl, np.array([[np.nan, np.nan], [3.0, 4.0], [5.0, 6.0], [7.0, np.nan]])
        )
        da2_int = np.array([[0, 1], [2, 3], [4, 5], [6, 2]])
        da2_fvint = 2
        replace_fillvals(da2_int, da2_fvint, "da2_int_scalarfill", "integer")
        assert_array_equal(da2_int, np.array([[0, 1], [0, 3], [4, 5], [6, 0]]))

    def test_print_servers(self):
        with self.assertLogs(level="ERROR") as log:
            hapi(trange=["2003-10-20", "2003-11-30"])
            self.assertIn(
                "No server specified; example servers include:", log.output[0]
            )

    def test_return_catalog(self):
        id_list = hapi(
            server="https://cdaweb.gsfc.nasa.gov/hapi", catalog=True, quiet=True
        )
        self.assertTrue("MMS1_EDI_BRST_L2_EFIELD" in id_list)

    def test_dataset_not_specified(self):
        # dataset not specified
        with self.assertLogs(level="ERROR") as log:
            h_vars = hapi(
                trange=["2003-10-20", "2003-11-30"],
                server="https://cdaweb.gsfc.nasa.gov/hapi",
            )
            self.assertIsNone(h_vars)
            self.assertIn(
                "Error, no dataset specified; please see the catalog for a list of available data sets.",
                log.output[0],
            )

    def test_trange_not_specified(self):
        # trange not specified
        with self.assertLogs(level="ERROR") as log:
            h_vars = hapi(
                dataset="OMNI_HRO2_1MIN", server="https://cdaweb.gsfc.nasa.gov/hapi"
            )
            self.assertIsNone(h_vars)
            self.assertIn("Error, no trange specified", log.output[0])

    def test_cdaweb_mms_spec(self):
        h_vars = hapi(
            trange=["2019-10-16", "2019-10-17"],
            server="https://cdaweb.gsfc.nasa.gov/hapi",
            dataset="MMS4_EDP_SRVY_L2_HFESP",
        )
        self.assertTrue(h_vars)

    def test_cdaweb_omni(self):
        del_data()
        h_vars = hapi(
            trange=["2003-10-20", "2003-11-30"],
            server="https://cdaweb.gsfc.nasa.gov/hapi",
            dataset="OMNI_HRO2_1MIN",
        )
        self.assertTrue(h_vars)
        self.assertTrue(data_exists("BX_GSE"))
        self.assertTrue(data_exists("BY_GSE"))
        self.assertTrue(data_exists("BZ_GSE"))

    def test_string_time(self):
        del_data()
        server = "https://supermag.jhuapl.edu/hapi"
        dataset = "ttb/baseline_all/PT1M/XYZ"
        start = "2020-05-10T00:00Z"
        stop = "2020-05-14T00:00Z"
        parameters = ""
        param_list = hapi(
            trange=[start, stop], server=server, dataset=dataset, parameters=parameters
        )
        self.assertTrue("mlt" in param_list)
        self.assertTrue(data_exists("mlt"))


if __name__ == "__main__":
    unittest.main()



import cdflib
import fnmatch

import numpy as np

from copy import deepcopy
from pyspedas.tplot_tools import get_data, store_data, options, clip, ylim, zlim
from pyspedas import tnames

from ....satellite.erg.load import load
from ....satellite.erg.get_gatt_ror import get_gatt_ror

from .get_sphcntr import get_sphcntr

"""
;Internal routine to get the table of the pixel
;centers from the table of the pixel corners.
"""


def get_pixel_cntr(tbl_array):
    dim_tuple = tbl_array.shape
    rgmax = dim_tuple[0] - 1
    azmax = dim_tuple[1] - 1
    cnttbl = np.zeros(shape=(rgmax, azmax, 2))
    for i in range(rgmax):
        for j in range(azmax):
            axis_0_indices_array = np.repeat(
                np.array([[i, i + 1, i + 1, i]]).T, 4, 1
            ).T.reshape(16)
            axis_1_indices_array = np.array([j] * 8 + [j + 1] * 8)
            lonarr = tbl_array[
                tuple([axis_0_indices_array, axis_1_indices_array, 0])
            ].reshape(4, 4)
            latarr = tbl_array[
                tuple([axis_0_indices_array, axis_1_indices_array, 1])
            ].reshape(4, 4)
            pos_array = get_sphcntr(latarr, lonarr)
            cnttbl[i, j, 1] = pos_array[0]
            cnttbl[i, j, 0] = pos_array[1]
    return cnttbl


from typing import List, Optional, Union


def sd_fit(
    trange: List[str] = ["2018-10-18/00:00:00", "2018-10-18/02:00:00"],
    suffix: str = "",
    site: Union[str, List[str]] = "all",
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    compact: bool = False,
    force_download: bool = False,
) -> List[str]:
    """
    Load SuperDARN data from ERG Science Center

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2020-08-01', '2020-08-02']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load. Valid values:  'ade', 'adw', 'bks', 'bpk',
            'cly', 'cve', 'cvw', 'dce', 'fhe', 'fhw', 'fir', 'gbr', 'hal', 'han', 'hok',
            'hkw', 'inv', 'kap', 'ker', 'kod', 'ksr', 'mcm', 'pgr', 'pyk', 'rkn', 'san',
            'sas', 'sps', 'sto', 'sye', 'sys', 'tig', 'unw', 'wal', 'zho', 'lyr', 'all
            Default: ['all']

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    compact: bool
            If True, leave only minimal set of the variables. Default: False

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------

    Examples
    ________
    >>> import pyspedas
    >>> sd_vars=pyspedas.projects.erg.sd_fit(trange=['2018-10-14/00:00:00','2018-10-14/02:00:00'],site='ade')
    >>> print(sd_vars)

    """

    valid_sites = [
        "ade",
        "adw",
        "bks",
        "bpk",
        "cly",
        "cve",
        "cvw",
        "dce",
        "fhe",
        "fhw",
        "fir",
        "gbr",
        "hal",
        "han",
        "hok",
        "hkw",
        "inv",
        "kap",
        "ker",
        "kod",
        "ksr",
        "mcm",
        "pgr",
        "pyk",
        "rkn",
        "san",
        "sas",
        "sps",
        "sto",
        "sye",
        "sys",
        "tig",
        "unw",
        "wal",
        "zho",
        "lyr",
    ]

    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(" ")
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if "all" in site_code:
        site_code = valid_sites
    site_code = list(set(site_code).intersection(valid_sites))


    new_cdflib = False
    if cdflib.__version__ > "0.4.9":
        new_cdflib = True
    else:
        new_cdflib = False

    if notplot:
        loaded_data = {}
    else:
        loaded_data = []

    for site_input in site_code:
        prefix = "sd_" + site_input + "_"
        file_res = 3600.0 * 24.0
        pathformat = (
            "ground/radar/sd/fitacf/"
            + site_input
            + "/%Y/sd_fitacf_l2_"
            + site_input
            + "_%Y%m%d*.cdf"
        )

        loaded_data_temp = load(
            pathformat=pathformat,
            file_res=file_res,
            trange=trange,
            prefix=prefix,
            suffix=suffix,
            get_support_data=get_support_data,
            varformat=varformat,
            downloadonly=downloadonly,
            notplot=notplot,
            time_clip=time_clip,
            no_update=no_update,
            uname=uname,
            passwd=passwd,
            force_download=force_download,
        )
        if notplot:
            loaded_data.update(loaded_data_temp)
        else:
            loaded_data += loaded_data_temp
        if (len(loaded_data_temp) > 0) and ror:
            try:
                gatt = get_gatt_ror(downloadonly, loaded_data)
                print("############## RULES OF THE ROAD ################")
                print(gatt["Rules_of_use"])
                print("############## RULES OF THE ROAD ################")
            except:
                print("printing PI info and rules of the road was failed")

        if (not downloadonly) and (not notplot) and (len(loaded_data_temp) > 0):
            t_plot_name_list = tnames(
                [prefix + "pwr*", prefix + "spec*", prefix + "vlos*"]
            )
            t_plot_name_list = list(set(t_plot_name_list).intersection(loaded_data))
            for t_plot_name in t_plot_name_list:
                clip(t_plot_name, -9000, 9000)

            t_plot_name_list = tnames([prefix + "elev*"])
            t_plot_name_list = list(set(t_plot_name_list).intersection(loaded_data))
            if len(t_plot_name_list) > 5:
                for t_plot_name in t_plot_name_list:
                    clip(t_plot_name, -9000, 9000)

            azim_no_name_list = list(
                set(tnames(prefix + "*azim_no_?" + suffix)).intersection(loaded_data)
            )
            number_string_list = []
            for azim_no_name in azim_no_name_list:
                number_string_list.append(azim_no_name.split("_")[4][0])

            for number_string in number_string_list:
                site_input_upper = site_input.upper()
                #  ;Set labels for some tplot variables
                options(
                    prefix + "pwr_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "pwr_" + number_string + suffix,
                    "ztitle",
                    "Backscatter power [dB]",
                )
                options(
                    prefix + "pwr_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "pwr_err_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "pwr_err_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "pwr_err_" + number_string + suffix,
                    "ztitle",
                    "power err [dB]",
                )
                options(
                    prefix + "spec_width_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "spec_width_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "spec_width_" + number_string + suffix,
                    "ztitle",
                    "Spec. width [m/s]",
                )
                options(
                    prefix + "spec_width_err_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "spec_width_err_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "spec_width_err_" + number_string + suffix,
                    "ztitle",
                    "Spec. width err [m/s]",
                )

                if not prefix + "vlos_" + number_string + suffix in loaded_data:
                    vlos_notplot_dictionary = load(
                        pathformat=pathformat,
                        file_res=file_res,
                        trange=trange,
                        prefix=prefix,
                        suffix=suffix,
                        get_support_data=get_support_data,
                        varformat="vlos_" + number_string,
                        downloadonly=downloadonly,
                        notplot=True,
                        time_clip=time_clip,
                        no_update=no_update,
                        uname=uname,
                        passwd=passwd,
                    )
                    vlos_tplot_name = prefix + "vlos_" + number_string + suffix
                    if len(vlos_notplot_dictionary) > 0:
                        store_data(
                            vlos_tplot_name,
                            data={
                                "x": vlos_notplot_dictionary[vlos_tplot_name]["x"],
                                "y": vlos_notplot_dictionary[vlos_tplot_name]["y"],
                                "v1": vlos_notplot_dictionary[vlos_tplot_name]["v"],
                                "v2": np.arange(
                                    vlos_notplot_dictionary[vlos_tplot_name]["y"].shape[
                                        2
                                    ]
                                ),
                            },
                            attr_dict={
                                "CDF": vlos_notplot_dictionary[vlos_tplot_name]["CDF"]
                            },
                        )

                        clip(vlos_tplot_name, -9000, 9000)
                        options(vlos_tplot_name, "spec", 1)
                        loaded_data.append(vlos_tplot_name)
                options(
                    prefix + "vlos_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "vlos_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "vlos_" + number_string + suffix,
                    "ztitle",
                    "Doppler velocity [m/s]",
                )

                options(
                    prefix + "vlos_err_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "vlos_err_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "vlos_err_" + number_string + suffix,
                    "ztitle",
                    "Vlos err [m/s]",
                )
                if (
                    prefix + "elev_angle_" + number_string + suffix in loaded_data
                ):  # need to get_support_data=True
                    options(
                        prefix + "elev_angle_" + number_string + suffix,
                        "ytitle",
                        site_input_upper + "\nall beams",
                    )
                    options(
                        prefix + "elev_angle_" + number_string + suffix,
                        "ysubtitle",
                        "[range gate]",
                    )
                    options(
                        prefix + "elev_angle_" + number_string + suffix,
                        "ztitle",
                        "Elev. angle [deg]",
                    )
                options(
                    prefix + "echo_flag_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "echo_flag_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "echo_flag_" + number_string + suffix,
                    "ztitle",
                    "1: iono. echo",
                )
                options(
                    prefix + "quality_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "quality_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "quality_" + number_string + suffix, "ztitle", "quality"
                )
                options(
                    prefix + "quality_flag_" + number_string + suffix,
                    "ytitle",
                    site_input_upper + "\nall beams",
                )
                options(
                    prefix + "quality_flag_" + number_string + suffix,
                    "ysubtitle",
                    "[range gate]",
                )
                options(
                    prefix + "quality_flag_" + number_string + suffix,
                    "ztitle",
                    "quality flg",
                )

                # ;Split vlos_? tplot variable into 3 components
                get_data_vlos = get_data(prefix + "vlos_" + number_string + suffix)
                if get_data_vlos is not None:
                    if get_data_vlos[1].ndim >= 3:
                        get_metadata_vlos = get_data(
                            prefix + "vlos_" + number_string + suffix, metadata=True
                        )
                        store_data(
                            prefix + "vnorth_" + number_string + suffix,
                            data={
                                "x": get_data_vlos[0],
                                "y": get_data_vlos[1][:, :, 0],
                                "v": get_data_vlos[2],
                            },
                            attr_dict=get_metadata_vlos,
                        )
                        options(
                            prefix + "vnorth_" + number_string + suffix,
                            "ztitle",
                            "LOS V Northward [m/s]",
                        )
                        loaded_data.append(prefix + "vnorth_" + number_string + suffix)
                        if get_data_vlos[1].shape[2] >= 1:
                            store_data(
                                prefix + "veast_" + number_string + suffix,
                                data={
                                    "x": get_data_vlos[0],
                                    "y": get_data_vlos[1][:, :, 1],
                                    "v": get_data_vlos[2],
                                },
                                attr_dict=get_metadata_vlos,
                            )
                            options(
                                prefix + "veast_" + number_string + suffix,
                                "ztitle",
                                "LOS V Eastward [m/s]",
                            )
                            loaded_data.append(
                                prefix + "veast_" + number_string + suffix
                            )
                            if get_data_vlos[1].shape[2] >= 2:
                                store_data(
                                    prefix + "vlos_" + number_string + suffix,
                                    data={
                                        "x": get_data_vlos[0],
                                        "y": get_data_vlos[1][:, :, 2],
                                        "v": get_data_vlos[2],
                                    },
                                    attr_dict=get_metadata_vlos,
                                )
                                options(
                                    prefix + "vlos_" + number_string + suffix,
                                    "ztitle",
                                    "LOS Doppler vel. [m/s]",
                                )

                    # ;Combine iono. echo and ground echo for vlos
                    v_var_names = ["vlos_", "vnorth_", "veast_"]
                    flag_data = get_data(prefix + "echo_flag_" + number_string + suffix)
                    for v_var in v_var_names:
                        v_var_data = get_data(prefix + v_var + number_string + suffix)
                        if v_var_data is not None:
                            v_var_metadata = get_data(
                                prefix + v_var + number_string + suffix, metadata=True
                            )
                            g_data_y = np.where(
                                flag_data[1] == 1.0, np.nan, v_var_data[1]
                            )
                            v_var_data_y = np.where(
                                flag_data[1] != 1.0, np.nan, v_var_data[1]
                            )
                            max_rg = np.nanmax(v_var_data[2]) + 1
                            store_data(
                                prefix + v_var + "iscat_" + number_string + suffix,
                                data={
                                    "x": v_var_data[0],
                                    "y": v_var_data_y,
                                    "v": v_var_data[2],
                                },
                                attr_dict=v_var_metadata,
                            )
                            options(
                                prefix + v_var + "iscat_" + number_string + suffix,
                                "ytitle",
                                " ",
                            )
                            options(
                                prefix + v_var + "iscat_" + number_string + suffix,
                                "ysubtitle",
                                " ",
                            )
                            options(
                                prefix + v_var + "iscat_" + number_string + suffix,
                                "ztitle",
                                " ",
                            )
                            options(
                                prefix + v_var + "iscat_" + number_string + suffix,
                                "spec",
                                1,
                            )
                            loaded_data.append(
                                prefix + v_var + "iscat_" + number_string + suffix
                            )
                            metadata_for_gscat = deepcopy(v_var_metadata)
                            metadata_for_gscat["plot_options"]["extras"][
                                "fill_color"
                            ] = 5  # options like, 'fill_color:5' in IDL, have not implemented.
                            store_data(
                                prefix + v_var + "gscat_" + number_string + suffix,
                                data={
                                    "x": v_var_data[0],
                                    "y": g_data_y,
                                    "v": v_var_data[2],
                                },
                                attr_dict=metadata_for_gscat,
                            )
                            options(
                                prefix + v_var + "gscat_" + number_string + suffix,
                                "ytitle",
                                " ",
                            )
                            options(
                                prefix + v_var + "gscat_" + number_string + suffix,
                                "ysubtitle",
                                " ",
                            )
                            options(
                                prefix + v_var + "gscat_" + number_string + suffix,
                                "ztitle",
                                " ",
                            )
                            options(
                                prefix + v_var + "gscat_" + number_string + suffix,
                                "spec",
                                1,
                            )
                            loaded_data.append(
                                prefix + v_var + "gscat_" + number_string + suffix
                            )
                            store_data(
                                prefix + v_var + "bothscat_" + number_string + suffix,
                                data=[
                                    prefix + v_var + "iscat_" + number_string + suffix,
                                    prefix + v_var + "gscat_" + number_string + suffix,
                                ],
                            )
                            options(
                                prefix + v_var + "bothscat_" + number_string + suffix,
                                "yrange",
                                [0, max_rg],
                            )
                            loaded_data.append(
                                prefix + v_var + "bothscat_" + number_string + suffix
                            )
                            """
                            Currently, '*iscat_*' and '*bothscat_*' are almost same plot outputs.
                            Because, options like, 'fill_color:5' of IDL for '*gscat_*' have not implemented.
                            """

                # ;Set the z range explicitly for some tplot variables
                zlim(prefix + "pwr_" + number_string + suffix, 0.0, 30.0)
                zlim(prefix + "pwr_err_" + number_string + suffix, 0.0, 30.0)
                zlim(prefix + "spec_width_" + number_string + suffix, 0.0, 200.0)
                zlim(prefix + "spec_width_err_" + number_string + suffix, 0.0, 300.0)

                # zlim for '*vlos_*scat_*'
                t_names_raw = tnames(prefix + "vlos_*scat_" + number_string + suffix)
                t_names_remove_space = [t_name.split(" ")[0] for t_name in t_names_raw]
                t_plot_name_list = list(
                    set(t_names_remove_space).intersection(loaded_data)
                )
                for t_plot_name in t_plot_name_list:
                    zlim(t_plot_name, -400.0, 400.0)

                # zlim for '*vnorth_*scat_*'
                t_names_raw = tnames(prefix + "vnorth_*scat_" + number_string + suffix)
                t_names_remove_space = [t_name.split(" ")[0] for t_name in t_names_raw]
                t_plot_name_list = list(
                    set(t_names_remove_space).intersection(loaded_data)
                )
                for t_plot_name in t_plot_name_list:
                    zlim(t_plot_name, -400.0, 400.0)

                # zlim for '*veast_*scat_*'
                t_names_raw = tnames(prefix + "veast_*scat_" + number_string + suffix)
                t_names_remove_space = [t_name.split(" ")[0] for t_name in t_names_raw]
                t_plot_name_list = list(
                    set(t_names_remove_space).intersection(loaded_data)
                )
                for t_plot_name in t_plot_name_list:
                    zlim(t_plot_name, -400.0, 400.0)

                zlim(prefix + "vlos_err_" + number_string + suffix, 0.0, 300.0)

                # ;Fill values --> NaN
                get_data_vars_pwr = get_data(prefix + "pwr_" + number_string + suffix)
                if get_data_vars_pwr is not None:
                    pwr_y = deepcopy(get_data_vars_pwr[1])
                    indices_array_tuple = np.where(np.isfinite(pwr_y) == False)
                    var_name_list = ["echo_flag_", "quality_", "quality_flag_"]
                    for var_name in var_name_list:
                        t_plot_name = prefix + var_name + number_string + suffix
                        get_data_vars = get_data(t_plot_name)
                        get_metadata_vars = get_data(t_plot_name, metadata=True)
                        if get_data_vars is not None:
                            val_array = deepcopy(get_data_vars[1].astype(np.float64))
                            val_array[indices_array_tuple] = np.nan
                            store_data(
                                t_plot_name,
                                data={
                                    "x": get_data_vars[0],
                                    "y": val_array,
                                    "v": get_data_vars[2],
                                },
                                attr_dict=get_metadata_vars,
                            )

                # ;Reassign scan numbers for the combined data
                if (
                    prefix + "scanstartflag_" + number_string + suffix in loaded_data
                ) and (
                    prefix + "scanno_" + number_string + suffix in loaded_data
                ):  # need to get_support_data=True
                    t_plot_name = prefix + "scanstartflag_" + number_string + suffix
                    scanstartflag_data = get_data(t_plot_name)
                    if scanstartflag_data is not None:
                        scflg = abs(scanstartflag_data[1])
                        try:
                            scno = np.full(
                                shape=scflg.shape, fill_value=-1, dtype=np.int64
                            )
                        except:
                            scno = np.full(
                                shape=scflg.shape, fill_value=-1, dtype=np.int32
                            )
                        scno_t = 0
                        scno[0] = scno_t
                        gt_1_indices_array = np.where(scflg > 0)[0]
                        for i in range(gt_1_indices_array.size - 1):
                            scno[gt_1_indices_array[i] : gt_1_indices_array[i + 1]] = i
                        scno[gt_1_indices_array[i + 1] :] = i + 1
                        t_plot_name = prefix + "scanno_" + number_string + suffix
                        get_data_var_scanno = get_data(t_plot_name)
                        if get_data_var_scanno is not None:
                            get_metadata_var_scanno = get_data(
                                t_plot_name, metadata=True
                            )
                            store_data(
                                t_plot_name,
                                data={"x": get_data_var_scanno[0], "y": scno},
                                attr_dict=get_metadata_var_scanno,
                            )

            """
            ;Load the position table(s) ;;;;;;;;;;;;;;;;;;
            ;Currently supports SD fitacf CDFs containing up to 4 pos. tables.
            """
            tbllist = [
                "tbl_0",
                "tbl_1",
                "tbl_2",
                "tbl_3",
                "tbl_4",
                "tbl_5",
                "tbl_6",
                "tbl_7",
                "tbl_8",
                "tbl_9",
            ]
            timelist = [
                "time_0",
                "time_1",
                "time_2",
                "time_3",
                "time_4",
                "time_5",
                "time_6",
                "time_7",
                "time_8",
                "time_9",
            ]

            get_metadata_vars = get_data(loaded_data[-1], metadata=True)
            if get_metadata_vars is not None:
                datfiles = deepcopy(get_metadata_vars["CDF"]["FILENAME"])
                if type(datfiles) is str:
                    datfiles = [datfiles]
                position_tbl_dictionary = {}
                for i in range(10):
                    position_tbl_dictionary[str(i)] = {
                        "time_input": [],
                        "tbl_input": [],
                        "cnttbl_input": [],
                    }
                if len(datfiles) > 0:
                    for file_name in datfiles:
                        cdf_file = cdflib.CDF(file_name)
                        cdf_info = cdf_file.cdf_info()
                        if new_cdflib:
                            all_cdf_variables = cdf_info.rVariables + cdf_info.zVariables
                        else:
                            all_cdf_variables = cdf_info["rVariables"] + cdf_info["zVariables"]

                        timevn = fnmatch.filter(all_cdf_variables, "Epoch_?")
                        ptblvn = fnmatch.filter(all_cdf_variables, "position_tbl_?")
                        timevn.sort()
                        ptblvn.sort()
                        for j in range(len(ptblvn)):
                            tv_name = timevn[j]
                            stblno = tv_name.split("_")[-1]
                            pv_name = ptblvn[j]
                            time_array = cdf_file.varget(tv_name)
                            tbl_array = cdf_file.varget(pv_name)
                            cnttbl = get_pixel_cntr(tbl_array)
                            position_tbl_dictionary[stblno]["time_input"] += [
                                time_array[0],
                                time_array[-1],
                            ]
                            dim_tuple = tbl_array.shape
                            tbl2_array = tbl_array.reshape(
                                1, dim_tuple[0], dim_tuple[1], dim_tuple[2]
                            )
                            cnttbl2_array = cnttbl.reshape(
                                1, dim_tuple[0] - 1, dim_tuple[1] - 1, dim_tuple[2]
                            )
                            if len(position_tbl_dictionary[stblno]["tbl_input"]) == 0:
                                position_tbl_dictionary[stblno][
                                    "tbl_input"
                                ] = np.concatenate([tbl2_array, tbl2_array], axis=0)
                            else:
                                position_tbl_dictionary[stblno][
                                    "tbl_input"
                                ] = np.concatenate(
                                    [
                                        position_tbl_dictionary[stblno]["tbl_input"],
                                        tbl2_array,
                                        tbl2_array,
                                    ],
                                    axis=0,
                                )
                            if (
                                len(position_tbl_dictionary[stblno]["cnttbl_input"])
                                == 0
                            ):
                                position_tbl_dictionary[stblno][
                                    "cnttbl_input"
                                ] = np.concatenate(
                                    [cnttbl2_array, cnttbl2_array], axis=0
                                )
                            else:
                                position_tbl_dictionary[stblno][
                                    "cnttbl_input"
                                ] = np.concatenate(
                                    [
                                        position_tbl_dictionary[stblno]["cnttbl_input"],
                                        cnttbl2_array,
                                        cnttbl2_array,
                                    ],
                                    axis=0,
                                )

                    for t_plot_suffix_number in position_tbl_dictionary.keys():
                        if (
                            len(
                                position_tbl_dictionary[t_plot_suffix_number][
                                    "time_input"
                                ]
                            )
                            >= 2
                        ):
                            input_tplot_time_array = (
                                np.array(
                                    position_tbl_dictionary[t_plot_suffix_number][
                                        "time_input"
                                    ]
                                )
                                / 1000.0
                                - 719528.0 * 24.0 * 3600.0
                            )
                            t_plot_name = (
                                prefix + "position_tbl_" + t_plot_suffix_number + suffix
                            )
                            store_data(
                                t_plot_name,
                                data={
                                    "x": input_tplot_time_array,
                                    "y": position_tbl_dictionary[t_plot_suffix_number][
                                        "tbl_input"
                                    ],
                                },
                            )
                            loaded_data.append(t_plot_name)
                            t_plot_name = (
                                prefix
                                + "positioncnt_tbl_"
                                + t_plot_suffix_number
                                + suffix
                            )
                            store_data(
                                t_plot_name,
                                data={
                                    "x": input_tplot_time_array,
                                    "y": position_tbl_dictionary[t_plot_suffix_number][
                                        "cnttbl_input"
                                    ],
                                },
                            )
                            loaded_data.append(t_plot_name)

    if compact:  # ;Leave only minimal set of the variables if compact=True.
        search_var_list = [
            "*cpid*",
            "*channel*",
            "*int_time*",
            "*azim_no*",
            "*pwr_err*",
            "*spec_width_err*",
            "*vlos_err*",
            "*elev_angle*",
            "*elev_angle_err*",
            "*phi0*",
            "*phi0_err*",
            "*echo_flag*",
            "*quality*",
            "*quality_flag*",
            "*scanno*",
            "*scanstartflag*",
            "*lagfr*",
            "*smsep*",
            "*nrang_max*",
            "*tfreq*",
            "*noise*",
            "*num_ave*",
            "*txpl*",
            "*vnorth*",
            "*veast*",
            "*vlos_bothscat*",
            "*vlos_iscat*",
            "*vlos_gscat*",
            "*vnorth_iscat*",
            "*vnorth_gscat*",
            "*vnorth_bothscat*",
            "*veast_iscat*",
            "*veast_gscat*",
            "*veast_bothscat*",
            "*position_tbl*",
            "*positioncnt_tbl*",
        ]
        delete_tplot_name_list = list(
            set(tnames(search_var_list)).intersection(loaded_data)
        )
        if len(delete_tplot_name_list) > 0:
            store_data(delete_tplot_name_list, delete=True)
            loaded_data = list(set(loaded_data).difference(delete_tplot_name_list))

    return loaded_data



"""
Apply a wavelet transformation to every component of a tplot variable.

Notes
-----
Similar to wav_data.pro in IDL SPEDAS.
For pywavelets library, see:
    https://pywavelets.readthedocs.io/en/latest/ref/cwt.html
For an example, see:
    http://spedas.org/wiki/index.php?title=Wavelet

"""
import logging
import numpy as np
import pywt
import pyspedas
from pyspedas.tplot_tools import options, store_data

def idl_wavelet_scales(n, dt, w0=2*np.pi, dj=None):
    """
    Reproduce the IDL/T&C logic for generating wavelet scales and frequencies.

    Parameters
    ----------
    n : int
        Number of samples in the time series.
    dt : float
        Sampling interval (in seconds).
    w0 : float
        Central frequency (default 2*pi for Morlet).
    dj : float or None
        Scale resolution step. Default: (1/8)*(2*pi/w0)

    Returns
    -------
    scales : ndarray
        Array of wavelet scales.
    freqs : ndarray
        Array of corresponding Fourier frequencies (Hz).
    periods : ndarray
        Fourier periods corresponding to each scale.
    """
    if dj is None:
        dj = (1/8) * (2*np.pi / w0)

    # Period range: [Nyquist period, 5% of total duration]
    nyquist_period = 2 * dt
    max_period = 0.05 * n * dt
    prange = [nyquist_period, max_period]

    # Scale range from period
    fourier_factor = (w0 + np.sqrt(2 + w0**2)) / (4 * np.pi)
    srange = np.array(prange) * fourier_factor

    # Number of scales
    j = int(np.floor(np.log2(srange[1] / srange[0]) / dj))
    if j <= 0:
        raise ValueError("Too few data points for wavelet analysis")

    # Log-spaced scales
    scales = srange[0] * 2**(np.arange(j + 1) * dj)

    # Periods and frequencies
    periods = scales / fourier_factor
    freqs = 1.0 / periods

    return scales, freqs, periods


def wavelet(
    names,
    newname=None,
    suffix='_pow',
    wavename='morl',
    scales=None,
    method='fft',
    sampling_period=1.0):
    """
    Find the wavelet transformation of a tplot variable.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    newname: str/list of str, optional
        List of new names for tplot variables.
        Default: If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply.
        Default: '_pow'.
    wavename: str, optional
        The name of the continuous wavelet function to apply.
        Examples: 'gaus1', 'morl', 'cmorlB-C'.
        Default: 'morl'
    scales: list of float, optional
        The wavelet scales to use.
        Default: None
    method: str, optional
        Either ‘fft’ for  frequency domain convolution,
        or 'conv' for numpy.convolve.
        Default: 'fft'
    sampling_period: float, optional
        The sampling period for the frequencies output.
        Default: 1.0

    Returns
    -------
    A list of tplot variables that contain the wavelet power.

    Example
    -------
        >>> import numpy as np
        >>> import pyspedas
        >>> from pyspedas import time_float
        >>> from pyspedas.analysis.wavelet import wavelet
        >>> # Create a tplot variable that contains a wave.
        >>> t = np.arange(4000.)
        >>> y = np.sin(2*np.pi*t/32.)
        >>> y2 = np.sin(2*np.pi*t/64.)
        >>> y[1000:3000] = y2[1000:3000]
        >>> var = 'sin_wav'
        >>> time = time_float('2010-01-01') + 10*t
        >>> pyspedas.store_data(var, data={'x':time, 'y':y})
        >>> # Gaussian Derivative wavelets transformation.
        >>> powervar = wavelet(var, wavename='gaus1')
        >>> pvar = powervar[0]
        >>> # Define plotting parameters and plot.
        >>> pyspedas.options(pvar, 'colormap', 'jet')
        >>> pyspedas.ylim(pvar, 0.001, 0.1)
        >>> pyspedas.options(pvar, 'ylog', True)
        >>> pyspedas.options(pvar, 'ytitle', pvar)
        >>> pyspedas.tplot([var, pvar])
    """

    varnames = pyspedas.split_vec(names)
    powervar = []

    if len(varnames) < 1:
        logging.error('wavelet error: No pytplot names were provided.')
        return

    for i, old in enumerate(varnames):
        old = varnames[i]

        if (newname is not None) and (len(newname) == len(varnames)):
            new = newname[i]
        else:
            new = old + suffix

        alldata = pyspedas.get_data(old)
        time = alldata[0]
        len_time = len(time)
        data = alldata[1]

        if len_time < 2:
            logging.error('wavelet error: Not enought data points for ' + old)
            continue

        if scales is None:
            n = len_time
            # This is just the first sample difference.  We should probably check that the dt values
            # are consistent throughout, and if not, interpolate onto a regular grid

            dt = time[1] - time[0]

            scales, idl_freqs, idl_periods = idl_wavelet_scales(n, dt)
            sampling_period = 1.0


        coef, freqs = pywt.cwt(data, scales=scales, wavelet=wavename,
                               method=method, sampling_period=sampling_period)

        power = np.abs(coef)**2
        power = power.transpose()
        store_data(new, data={'x': time, 'y': power, 'v': freqs})
        options(new, 'spec', 1)
        powervar.append(new)

        logging.info('wavelet was applied to: ' + new)

    return powervar



import pyspedas
import copy
import logging


def interp_nan(tvar, newname=None, s_limit=None):
    """
    Interpolates the tplot variable through NaNs in the data. This is basically just a wrapper for xarray's interpolate_na function.

    .. note::
        This analysis routine assumes the data is no more than 2 dimensions. If there are more, they may become flattened!

    Parameters
    ----------
    tvar : str
        Name of tplot variable.
    newname : str
        Name of new tvar for added data. If not set, then the original tvar is replaced.
    s_limit : int or float, optional
        The maximum size of the gap in seconds to not interpolate over. I.e. if there are too many NaNs in a row, leave them there.

    Returns
    -------
    None

    Examples
    --------
    >>> import pyspedas
    >>> import numpy as np
    >>> pyspedas.store_data('e', data={'x':[2,5,8,11,14,17,21], 'y':[[np.nan,1,1],[np.nan,2,3],[4,np.nan,47],[4,np.nan,5],[5,5,99],[6,6,25],[7,np.nan,-5]]})
    >>> pyspedas.interp_nan('e','e_nonan',s_limit=5)

    """

    x = pyspedas.tplot_tools.data_quants[tvar].interpolate_na(dim="time", limit=s_limit)
    x.attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)

    if newname is None:
        pyspedas.tplot_tools.data_quants[tvar] = x
        x.name = tvar
    else:
        pyspedas.tplot_tools.data_quants[newname] = x
        pyspedas.tplot_tools.data_quants[newname].name = newname



from pyspedas.projects.mms.mec_ascii.mms_get_state_data import mms_get_state_data



def mms_load_state(trange=['2015-10-16', '2015-10-17'], probe='1', level='def',
    datatypes=['pos', 'vel'], no_update=False, pred_or_def=True, suffix=''):
    """
    This function loads the state (ephemeris and attitude) data from the ASCII files 
    into tplot variables
    
    Parameters
    ----------
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2015-10-16', '2015-10-17']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4'].
            Default: '1'

        level : str
            indicates level of data (options: 'def' (definitive), 'pred' (predicted)
            Default: 'def'

        datatypes : str or list of str
            Datatypes for state data to be loaded (options: 'pos', 'vel', 'spinras', 'spindec')
            Default: ['pos', 'vel']

        suffix: str
            The tplot variable names will be given this suffix.
            Default: None

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten
            Default: False

        pred_or_def: bool
            Load definitive or predicted (if definitive isn't available); defaults to True
            Default: True

    Returns
    --------
        list of str
            List of tplot variables created.

    Example
    -------

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> pos_data = pyspedas.projects.mms.mms_load_state(trange=['2015-10-16', '2015-10-17'], probe='1', datatypes='pos', level='def')
    >>> tplot('mms1_defeph_pos')


    """
    return mms_get_state_data(trange=trange, probe=probe, level=level, datatypes=datatypes,
        no_download=no_update, pred_or_def=pred_or_def, suffix=suffix)



import numpy as np
from pyspedas import tnames
from pyspedas import tcrossp
from pyspedas.tplot_tools import time_string
from pyspedas.tplot_tools import get_data, get_timespan, degap
from scipy import interpolate

from ...att.att import att


def erg_interpolate_att(erg_xxx_in=None, noload=False):
    """
    This function interpolates erg att data to match erg_xxx_in.

    Parameters:
        erg_xxx_in : str
            input tplot variable relating to ERG to be transformed

    Returns:
        output_dictionary : dict
            Dictionary which has below keys.
                spinperiod: output variable in which the interpolated data of ERG spin period is stored
                spinphase: output variable in which the interpolated data of ERG spin phase is stored
                sgix_j2000, sgiy_j2000, or sgiz_j2000: output interporated SGI axis vector for each component
                sgax_j2000, sgay_j2000, or sgaz_j2000: output interporated SGA axis vector for each component

    """
    if (erg_xxx_in is None) or (erg_xxx_in not in tnames()):
        print('inputted Tplot variable name is None, or not defined')
        return

    reload = not noload

    time_array = get_data(erg_xxx_in)[0]

    # Prepare some constants
    dtor = np.pi / 180.

    output_dictionary = {}

    # Load the attitude data
    if tnames('erg_att_sprate') == ['erg_att_sprate']:
        if reload:
            degap('erg_att_sprate', dt=8., margin=.5)
        sprate = get_data('erg_att_sprate')
        if sprate[0].min() > time_array.min() + 8. or sprate[0].max() < time_array.max() - 8.:
            tr = get_timespan(erg_xxx_in)
            if reload:
                att(trange=time_string([tr[0] - 60., tr[1] + 60.]))
    else:
        tr = get_timespan(erg_xxx_in)
        if reload:
            att(trange=time_string([tr[0] - 60., tr[1] + 60.]))

    # Interpolate spin period
    if reload:
        degap('erg_att_sprate', dt=8., margin=.5)
    sprate = get_data('erg_att_sprate')
    sper = 1. / (sprate[1] / 60.)
    sperInterp = np.interp(time_array, sprate[0], sper)
    spinperiod = {'x': time_array, 'y': sperInterp}
    output_dictionary['spinperiod'] = spinperiod

    # Interpolate spin phase
    if reload:
        degap('erg_att_spphase', dt=8., margin=.5)
    sphase = get_data('erg_att_spphase')
    if (sphase[0][0] <= time_array[0])\
            and (time_array[-1] <= sphase[0][-1]):
        ph_nn = interpolate.interp1d(
            sphase[0], sphase[1], kind="nearest")(time_array)
        dt = time_array - \
            interpolate.interp1d(sphase[0], sphase[0], kind="nearest")(time_array)
    elif (time_array[0] < sphase[0][0])\
            or (sphase[0][-1] < time_array)[-1]:
        ph_nn = interpolate.interp1d(
            sphase[0], sphase[1], kind="nearest", fill_value='extrapolate')(time_array)
        dt = time_array - \
            interpolate.interp1d(sphase[0], sphase[0], kind="nearest", fill_value='extrapolate')(time_array)

    per_nn = spinperiod['y']

    sphInterp = np.fmod(ph_nn + 360. * dt / per_nn, 360.)
    sphInterp = np.fmod(sphInterp + 360., 360.)
    spinphase = {'x': time_array, 'y': sphInterp}
    output_dictionary['spinphase'] = spinphase

    # Interporate SGI-Z axis vector
    if reload:
        degap('erg_att_izras', dt=8., margin=0.5)
        degap('erg_att_izdec', dt=8., margin=0.5)
    ras = get_data('erg_att_izras')
    dec = get_data('erg_att_izdec')
    time0 = ras[0]
    ras = ras[1]
    dec = dec[1]
    ez = np.cos((90. - dec) * dtor)
    ex = np.sin((90. - dec) * dtor) * np.cos(ras * dtor)
    ey = np.sin((90. - dec) * dtor) * np.sin(ras * dtor)
    ex_interp = np.interp(time_array, time0, ex)
    ey_interp = np.interp(time_array, time0, ey)
    ez_interp = np.interp(time_array, time0, ez)
    sgiz_j2000 = {'x': time_array, 'y': np.array(
        [ex_interp, ey_interp, ez_interp]).T}
    output_dictionary['sgiz_j2000'] = sgiz_j2000

    # Interporate SGA-X axis vector
    if reload:
        degap('erg_att_gxras', dt=8., margin=0.5)
        degap('erg_att_gxdec', dt=8., margin=0.5)
    ras = get_data('erg_att_gxras')
    dec = get_data('erg_att_gxdec')
    time0 = ras[0]
    ras = ras[1]
    dec = dec[1]
    ez = np.cos((90. - dec) * dtor)
    ex = np.sin((90. - dec) * dtor) * np.cos(ras * dtor)
    ey = np.sin((90. - dec) * dtor) * np.sin(ras * dtor)
    ex_interp = np.interp(time_array, time0, ex)
    ey_interp = np.interp(time_array, time0, ey)
    ez_interp = np.interp(time_array, time0, ez)
    sgax_j2000 = {'x': time_array, 'y': np.array(
        [ex_interp, ey_interp, ez_interp]).T}
    output_dictionary['sgax_j2000'] = sgax_j2000

    # Interporate SGA-Z axis vector
    if reload:
        degap('erg_att_gzras', dt=8., margin=0.5)
        degap('erg_att_gzdec', dt=8., margin=0.5)
    ras = get_data('erg_att_gzras')
    dec = get_data('erg_att_gzdec')
    time0 = ras[0]
    ras = ras[1]
    dec = dec[1]
    ez = np.cos((90. - dec) * dtor)
    ex = np.sin((90. - dec) * dtor) * np.cos(ras * dtor)
    ey = np.sin((90. - dec) * dtor) * np.sin(ras * dtor)
    ex_interp = np.interp(time_array, time0, ex)
    ey_interp = np.interp(time_array, time0, ey)
    ez_interp = np.interp(time_array, time0, ez)
    sgaz_j2000 = {'x': time_array, 'y': np.array(
        [ex_interp, ey_interp, ez_interp]).T}
    output_dictionary['sgaz_j2000'] = sgaz_j2000

    # Derive the other three axes (SGA-Y, SGI-X, SGI-Y)
    sgay = tcrossp(output_dictionary['sgaz_j2000']['y'],
                   output_dictionary['sgax_j2000']['y'], return_data=True)
    sgay_j2000 = {'x': time_array, 'y': sgay}
    output_dictionary['sgay_j2000'] = sgay_j2000

    sgiy = tcrossp(output_dictionary['sgiz_j2000']['y'],
                   output_dictionary['sgax_j2000']['y'], return_data=True)
    sgiy_j2000 = {'x': time_array, 'y': sgiy}
    output_dictionary['sgiy_j2000'] = sgiy_j2000

    sgix = tcrossp(output_dictionary['sgiy_j2000']['y'],
                   output_dictionary['sgiz_j2000']['y'], return_data=True)
    sgix_j2000 = {'x': time_array, 'y': sgix}
    output_dictionary['sgix_j2000'] = sgix_j2000

    return output_dictionary



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import datetime
import numpy as np
from dateutil.parser import parse
import pyspedas
import copy


def convert_tplotxarray_to_pandas_dataframe(name, no_spec_bins=False):
    import pandas as pd
    # This function is not final, and will presumably change in the future
    # This function sums over all dimensions except for the second non-time one.
    # This collapses many dimensions into just a single "energy bin" dimensions

    matrix = pyspedas.tplot_tools.data_quants[name].values
    while len(matrix.shape) > 3:
        matrix = np.nansum(matrix, 3)
    if len(matrix.shape) == 3:
        matrix = np.nansum(matrix, 1)
    return_data = pd.DataFrame(matrix)
    return_data = return_data.set_index(pd.Index(pyspedas.tplot_tools.data_quants[name].coords['time'].values))

    if no_spec_bins:
        return return_data

    if 'spec_bins' in pyspedas.tplot_tools.data_quants[name].coords:
        spec_bins = pd.DataFrame(pyspedas.tplot_tools.data_quants[name].coords['spec_bins'].values)
        if len(pyspedas.tplot_tools.data_quants[name].coords['spec_bins'].shape) == 1:
            spec_bins = spec_bins.transpose()
        else:
            spec_bins = spec_bins.set_index(pd.Index(pyspedas.tplot_tools.data_quants[name].coords['time'].values))

        return return_data, spec_bins

    return return_data



from .load import load


# Add the loader for MagEphem cdf files.
def magephem(trange=["2018-11-5", "2018-11-6"], probe="a", cadence="1min", coord="op77q", prefix="", suffix="", force_download=False, get_support_data=False, varformat=None, varnames=[], downloadonly=False, notplot=False, no_update=False, time_clip=False):  # 1 min or 5 min  # op77q, t89d, t89q, and ts04d
    """
    This function loads the Mag Ephemeris data for the Van Allen Probes
    The data is available from:
    https://spdf.gsfc.nasa.gov/pub/data/rbsp/rbspa/ephemeris

    Parameters
    ----------
        trange : list of str, default=['2018-11-5', '2018-11-6']
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str, default='a'
            Spacecraft probe name: 'a' or 'b'

        cadence : str, default='1min'
            Data cadence; options: '1min', '5min'

        coord : str, default='op77q'
            Data coordinate system. options: 'op77q', 't89d', 't89q', and 'ts04d'

        prefix : str, optional
            The tplot variable names will be given this prefix. By default, no prefix is added.

        suffix : str, optional
            Suffix for tplot variable names. By default, no suffix is added.

        force_download : bool, default=False
            Download file even if local version is more recent than server version.

        get_support_data : bool, default=False
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot. By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat : str, optional
            The file variable formats to load into tplot. Wildcard character
            "*" is accepted. By default, all variables are loaded in.

        varnames: list of str, optional
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool, default=False
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool, default=False
            Return the data in hash tables instead of creating tplot variables

        no_update: bool, default=False
            If set, only load data from your local cache

        time_clip: bool, default=False
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    -------
    tvars : dict or list
        List of created tplot variables or dict of data tables if notplot is True.

    Examples
    --------
    >>> magephem_vars = pyspedas.projects.rbsp.magephem(trange=['2018-11-5/10:00', '2018-11-5/15:00'], cadence='1min', coord='op77q', time_clip=True)
    >>> tplot(['L','Lsimple','Kp','Dst'])
    """
    return load(instrument="magephem", trange=trange, probe=probe, cadence=cadence, coord=coord, prefix=prefix, suffix=suffix, force_download=force_download, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

